r"""``sphobjinv`` *CLI path resolution module*.

``sphobjinv`` is a toolkit for manipulation and inspection of
Sphinx |objects.inv| files.

**Author**
    Brian Skinn (brian.skinn@gmail.com)

**File Created**
    19 Nov 2020

**Copyright**
    \(c) 2016-2026 Brian Skinn and community contributors

**Source Repository**
    https://github.com/bskinn/sphobjinv

**Documentation**
    https://sphobjinv.readthedocs.io/en/stable

**License**
    Code: `MIT License`_

    Docs & Docstrings: |CC BY 4.0|_

    See |license_txt|_ for full license terms.

**Members**

"""

import os

from sphobjinv.cli.parser import PrsConst


def resolve_inpath(in_path):
    """Resolve the input file, handling invalid values.

    Currently, only checks for existence and not-directory.

    Parameters
    ----------
    in_path

        |str| -- Path to desired input file

    Returns
    -------
    abs_path

        |str| -- Absolute path to indicated file

    Raises
    ------
    :exc:`FileNotFoundError`

        If a file is not found at the given path

    """
    # Path MUST be to a file, that exists
    if not os.path.isfile(in_path):
        raise FileNotFoundError("Indicated path is not a valid file")

    # Return the path as absolute
    return os.path.abspath(in_path)


def resolve_outpath(out_path, in_path, params):
    r"""Resolve the output location, handling mode-specific defaults.

    If the output path or basename are not specified, they are
    taken as the same as the input file. If the extension is
    unspecified, it is taken as the appropriate mode-specific value
    from DEF_OUT_EXT.

    If URL is passed, the input directory
    is taken to be :func:`os.getcwd` and the input basename
    is taken as DEF_BASENAME.

    Parameters
    ----------
    out_path

        |str| or |None| -- Output location provided by the user,
        or |None| if omitted

    in_path

        |str| -- For a local input file, its absolute path.
        For a URL, the (possibly truncated) URL text.

    params

        |dict| -- Parameters/values mapping from the active subparser

    Returns
    -------
    out_path

        |str| -- Absolute path to the target output file

    """
    mode = params[PrsConst.MODE]

    if params[PrsConst.URL] or in_path is None:
        in_fld = os.getcwd()
        in_fname = PrsConst.DEF_BASENAME
    else:
        in_fld, in_fname = os.path.split(in_path)

    if out_path:
        # Must check if the path entered is a folder
        if os.path.isdir(out_path):
            # Set just the folder and leave the name blank
            out_fld = out_path
            out_fname = None
        else:
            # Split appropriately
            out_fld, out_fname = os.path.split(out_path)

        # Output to same folder if unspecified
        if not out_fld:
            out_fld = in_fld

        # Use same base filename if not specified
        if not out_fname:
            out_fname = os.path.splitext(in_fname)[0] + PrsConst.DEF_OUT_EXT[mode]

        # Composite the full output path
        out_path = os.path.join(out_fld, out_fname)
    else:
        # No output location specified; use defaults
        out_fname = os.path.splitext(in_fname)[0] + PrsConst.DEF_OUT_EXT[mode]
        out_path = os.path.join(in_fld, out_fname)

    return out_path



r"""*Module for* ``sphobjinv`` *CLI* |Inventory| *writing*.

``sphobjinv`` is a toolkit for manipulation and inspection of
Sphinx |objects.inv| files.

**Author**
    Brian Skinn (brian.skinn@gmail.com)

**File Created**
    19 Nov 2020

**Copyright**
    \(c) 2016-2026 Brian Skinn and community contributors

**Source Repository**
    https://github.com/bskinn/sphobjinv

**Documentation**
    https://sphobjinv.readthedocs.io/en/stable

**License**
    Code: `MIT License`_

    Docs & Docstrings: |CC BY 4.0|_

    See |license_txt|_ for full license terms.

**Members**

"""

import json
import os
import sys

from sphobjinv.cli.parser import PrsConst
from sphobjinv.cli.paths import resolve_outpath
from sphobjinv.cli.ui import err_format, print_stderr, yesno_prompt
from sphobjinv.fileops import writebytes, writejson
from sphobjinv.zlib import compress


def write_plaintext(inv, path, *, expand=False, contract=False):
    """Write an |Inventory| to plaintext.

    Newlines are inserted in an OS-aware manner,
    based on the value of :data:`os.linesep`.

    Calling with both `expand` and `contract` as |True| is invalid.

    Parameters
    ----------
    inv

        |Inventory| -- Objects inventory to be written as plaintext

    path

        |str| -- Path to output file

    expand

        |bool| *(optional)* -- Generate output with any
        :data:`~sphobjinv.data.SuperDataObj.uri` or
        :data:`~sphobjinv.data.SuperDataObj.dispname`
        abbreviations expanded

    contract

        |bool| *(optional)* -- Generate output with abbreviated
        :data:`~sphobjinv.data.SuperDataObj.uri` and
        :data:`~sphobjinv.data.SuperDataObj.dispname` values

    Raises
    ------
    ValueError

        If both `expand` and `contract` are |True|

    """
    b_str = inv.data_file(expand=expand, contract=contract)
    writebytes(path, b_str.replace(b"\n", os.linesep.encode("utf-8")))


def write_zlib(inv, path, *, expand=False, contract=False):
    """Write an |Inventory| to zlib-compressed format.

       Calling with both `expand` and `contract` as |True| is invalid.

    Parameters
    ----------
    inv

        |Inventory| -- Objects inventory to be written zlib-compressed

    path

        |str| -- Path to output file

    expand

        |bool| *(optional)* -- Generate output with any
        :data:`~sphobjinv.data.SuperDataObj.uri` or
        :data:`~sphobjinv.data.SuperDataObj.dispname`
        abbreviations expanded

    contract

        |bool| *(optional)* -- Generate output with abbreviated
        :data:`~sphobjinv.data.SuperDataObj.uri` and
        :data:`~sphobjinv.data.SuperDataObj.dispname` values

    Raises
    ------
    ValueError

        If both `expand` and `contract` are |True|

    """
    b_str = inv.data_file(expand=expand, contract=contract)
    bz_str = compress(b_str)
    writebytes(path, bz_str)


def write_json(inv, path, params):
    """Write an |Inventory| to JSON.

    Writes output via
    :func:`fileops.writejson() <sphobjinv.fileops.writejson>`.

    Calling with both `expand` and `contract` as |True| is invalid.

    Parameters
    ----------
    inv

        |Inventory| -- Objects inventory to be written zlib-compressed

    path

        |str| -- Path to output file

    params

        dict -- `argparse` parameters

    Raises
    ------
    ValueError

        If both `params["expand"]` and `params["contract"]` are |True|

    """
    json_dict = inv.json_dict(
        expand=params[PrsConst.EXPAND], contract=params[PrsConst.CONTRACT]
    )

    if params.get(PrsConst.FOUND_URL, False):
        json_dict.update({"metadata": {PrsConst.URL: params[PrsConst.FOUND_URL]}})

    writejson(path, json_dict)


def write_stdout(inv, params):
    r"""Write the inventory contents to stdout.

    Parameters
    ----------
    inv

        |Inventory| -- Objects inventory to be written to stdout

    params

        dict -- `argparse` parameters

    Raises
    ------
    ValueError

        If both `params["expand"]` and `params["contract"]` are |True|

    """
    if params[PrsConst.MODE] == PrsConst.PLAIN:
        print(
            inv.data_file(
                expand=params[PrsConst.EXPAND], contract=params[PrsConst.CONTRACT]
            ).decode()
        )
    elif params[PrsConst.MODE] == PrsConst.JSON:
        json_dict = inv.json_dict(
            expand=params[PrsConst.EXPAND], contract=params[PrsConst.CONTRACT]
        )

        if params.get(PrsConst.FOUND_URL, False):
            json_dict.update({"metadata": {PrsConst.URL: params[PrsConst.FOUND_URL]}})

        print(json.dumps(json_dict))
    else:
        print_stderr("Error: Only plaintext and JSON can be emitted to stdout.", params)
        sys.exit(1)


def write_file(inv, in_path, params):
    r"""Write the inventory contents to a file on disk.

    Parameters
    ----------
    inv

        |Inventory| -- Objects inventory to be written to stdout

    in_path

        |str| -- For a local input file, its absolute path.
        For a URL, the (possibly truncated) URL text.

    params

        dict -- `argparse` parameters

    Raises
    ------
    ValueError

        If both `params["expand"]` and `params["contract"]` are |True|

    """
    mode = params[PrsConst.MODE]

    # Work up the output location
    try:
        out_path = resolve_outpath(params[PrsConst.OUTFILE], in_path, params)
    except Exception as e:  # pragma: no cover
        # This may not actually be reachable except in exceptional situations
        print_stderr("\nError while constructing output file path:", params)
        print_stderr(err_format(e), params)
        sys.exit(1)

    # If exists, must handle overwrite
    if os.path.isfile(out_path) and not params[PrsConst.OVERWRITE]:
        if params[PrsConst.INFILE] == "-":
            # If reading from stdin, just alert and don't overwrite
            print_stderr("\nFile exists. To overwrite, supply '-o'. Exiting...", params)
            sys.exit(0)
        # This could be written w/o nesting via elif, but would be harder to read.
        else:
            if not params[PrsConst.QUIET]:
                # If not a stdin read, confirm overwrite; or, just clobber if QUIET
                resp = yesno_prompt("File exists. Overwrite (Y/N)? ")
                if resp.lower() == "n":
                    print_stderr("\nExiting...", params)
                    sys.exit(0)

    # Write the output file
    try:
        if mode == PrsConst.ZLIB:
            write_zlib(
                inv,
                out_path,
                expand=params[PrsConst.EXPAND],
                contract=params[PrsConst.CONTRACT],
            )
        if mode == PrsConst.PLAIN:
            write_plaintext(
                inv,
                out_path,
                expand=params[PrsConst.EXPAND],
                contract=params[PrsConst.CONTRACT],
            )
        if mode == PrsConst.JSON:
            write_json(inv, out_path, params)
    except Exception as e:
        print_stderr("\nError during write of output file:", params)
        print_stderr(err_format(e), params)
        sys.exit(1)

    # Report success, if not QUIET
    print_stderr(
        "Conversion completed.\n"
        f"'{in_path if in_path else 'stdin'}' converted to '{out_path}' ({mode}).",
        params,
    )



r"""``sphobjinv`` *CLI UI functions*.

``sphobjinv`` is a toolkit for manipulation and inspection of
Sphinx |objects.inv| files.

**Author**
    Brian Skinn (brian.skinn@gmail.com)

**File Created**
    19 Nov 2020

**Copyright**
    \(c) 2016-2026 Brian Skinn and community contributors

**Source Repository**
    https://github.com/bskinn/sphobjinv

**Documentation**
    https://sphobjinv.readthedocs.io/en/stable

**License**
    Code: `MIT License`_

    Docs & Docstrings: |CC BY 4.0|_

    See |license_txt|_ for full license terms.

**Members**

"""

import sys

from sphobjinv.cli.parser import PrsConst


def print_stderr(thing, params, *, end="\n"):
    r"""Print `thing` to stderr if not in quiet mode.

    Quiet mode is indicated by the value at the QUIET key
    within `params`.

    Quiet mode is not implemented for the ":doc:`suggest </cli/suggest>`"
    CLI mode.

    Parameters
    ----------
    thing

        *any* -- Object to be printed

    params

        |dict| -- Parameters/values mapping from the active subparser

    end

        |str| -- String to append to printed content (default: ``\n``\ )

    """
    if (
        PrsConst.SUBPARSER_NAME not in params  # textconv
        or params[PrsConst.SUBPARSER_NAME][:2] == "su"  # suggest is never quiet
        or not params[PrsConst.QUIET]  # non-quiet convert
    ):
        print(thing, file=sys.stderr, end=end)


def err_format(exc):
    r"""Pretty-format an exception.

    Parameters
    ----------
    exc

        :class:`Exception` -- Exception instance to pretty-format

    Returns
    -------
    pretty_exc

        |str| -- Exception type and message formatted as
        |cour|\ '{type}: {message}'\ |/cour|

    """
    return f"{type(exc).__name__}: {str(exc)}"


def yesno_prompt(prompt):
    r"""Query user at `stdin` for yes/no confirmation.

    Uses :func:`input`, so will hang if used programmatically
    unless `stdin` is suitably mocked.

    The value returned from :func:`input` must satisfy either
    |cour|\ resp.lower() == 'n'\ |/cour| or
    |cour|\ resp.lower() == 'y'\ |/cour|,
    or else the query will be repeated *ad infinitum*.
    This function does **NOT** augment `prompt`
    to indicate the constraints on the accepted values.

    Parameters
    ----------
    prompt

        |str| -- Prompt to display to user that
        requests a 'Y' or 'N' response

    Returns
    -------
    resp

        |str| -- User response

    """
    resp = ""
    while not (resp.lower() == "n" or resp.lower() == "y"):
        resp = input(prompt)
    return resp



r"""*JSON schema to validate inventory dictionaries*.

This module is part of ``sphobjinv``,
a toolkit for manipulation and inspection of
Sphinx |objects.inv| files.

**Author**
    Brian Skinn (brian.skinn@gmail.com)

**File Created**
    7 Dec 2017

**Copyright**
    \(c) 2016-2026 Brian Skinn and community contributors

**Source Repository**
    https://github.com/bskinn/sphobjinv

**Documentation**
    https://sphobjinv.readthedocs.io/en/stable

**License**
    Code: `MIT License`_

    Docs & Docstrings: |CC BY 4.0|_

    See |license_txt|_ for full license terms.

**Members**

"""

# For jsonschema Draft 4.
# Schemas are defined with static field names as a versioning
# guarantee, instead of basing them dynamically on DataFields, etc.

# JSON dict schema
# Subschema for the inner data, both for clarity and to make it
# possible to satisfy flake8
subschema_json = {
    "name": {"type": "string"},
    "domain": {"type": "string"},
    "role": {"type": "string"},
    "priority": {"type": "string"},
    "uri": {"type": "string"},
    "dispname": {"type": "string"},
}

#: JSON schema for validating the |dict| forms of
#: Sphinx |objects.inv| inventories
#: as generated from or expected by
#: :class:`~sphobjinv.inventory.Inventory` classes.
json_schema = {
    "$schema": "https://json-schema.org/draft-04/schema#",
    "type": "object",
    "properties": {
        "project": {"type": "string"},
        "version": {"type": "string"},
        "count": {"type": "integer"},
        "metadata": {},
    },
    "patternProperties": {
        r"^\d+": {
            "type": "object",
            "properties": subschema_json,
            "additionalProperties": False,
            "required": list(subschema_json),
        }
    },
    "additionalProperties": False,
    "required": ["project", "version", "count"],
}



r"""``sphobjinv`` *module for CLI suggest functionality*.

``sphobjinv`` is a toolkit for manipulation and inspection of
Sphinx |objects.inv| files.

**Author**
    Brian Skinn (brian.skinn@gmail.com)

**File Created**
    20 Oct 2022

**Copyright**
    \(c) 2016-2026 Brian Skinn and community contributors

**Source Repository**
    https://github.com/bskinn/sphobjinv

**Documentation**
    https://sphobjinv.readthedocs.io/en/stable

**License**
    Code: `MIT License`_

    Docs & Docstrings: |CC BY 4.0|_

    See |license_txt|_ for full license terms.

**Members**

"""

import itertools as itt
import shutil
import sys
import urllib.parse as urlparse

from sphobjinv.cli.parser import PrsConst
from sphobjinv.cli.ui import print_stderr, yesno_prompt


def do_suggest(inv, params):
    r"""Perform the suggest call and output the results.

    Results are printed one per line.

    If neither INDEX nor SCORE is specified,
    the results are output without a header.
    If either or both are specified,
    the results are output in a lightweight tabular format.

    If the number of results exceeds
    SUGGEST_CONFIRM_LENGTH,
    the user will be queried whether to display
    all of the returned results
    unless ALL is specified.

    No QUIET option is available here, since
    a silent mode for suggestion output is nonsensical.

    Parameters
    ----------
    inv

        |Inventory| -- Inventory object to be output in the format
        indicated by MODE.

    params

        |dict| -- Parameters/values mapping from the active subparser

    """
    with_index = params[PrsConst.INDEX]
    with_score = params[PrsConst.SCORE]
    results = inv.suggest(
        params[PrsConst.SEARCH],
        thresh=params[PrsConst.THRESH],
        with_index=with_index,
        with_score=with_score,
    )

    print_divider(params)
    print_stderr_inferred_mapping(params)
    print_divider(params)

    print_stderr(f"Project: {inv.project}", params)
    print_stderr(f"Version: {inv.version}\n", params)

    print_stderr(f"{inv.count} objects in inventory.\n", params)
    print_divider(params)

    print_stderr_result_count(params, results)

    if not results:
        print_stderr("\nExiting...\n", params)
        sys.exit(0)

    # The query here for printing the full list only occurs in some
    # circumstances; see the function docstring.
    confirm_print_if_long_list(params, results)

    print_results_table(with_index, with_score, results, params)


def print_divider(params):
    """Print a visual divider to break up sections of the CLI output."""
    length = shutil.get_terminal_size().columns * 3 // 5
    print_stderr("-" * length + "\n", params)


def print_stderr_result_count(params, results):
    """Report the count of found objects from the suggest search."""
    if len(results) == 0:
        print_stderr(
            (
                "No results found with score at/above current threshold of "
                f"{params[PrsConst.THRESH]}."
            ),
            params,
        )
    else:
        print_stderr(
            (
                f"{len(results)} result"
                f"{'' if len(results) == 1 else 's'}"
                " found at/above current threshold of "
                f"{params[PrsConst.THRESH]}.\n"
            ),
            params,
        )


def confirm_print_if_long_list(params, results):
    """Check if the results list is too long and query user if to proceed.

    Skip the check if ``--all`` has been passed.

    Also skip the check if receiving data from ``stdin``, as a stream
    interaction here fouls the data flow...I forget exactly how.

    """
    if (
        len(results) > PrsConst.SUGGEST_CONFIRM_LENGTH
        and not params[PrsConst.ALL]
        and params[PrsConst.INFILE] != "-"
    ):
        resp = yesno_prompt(f"Display all {len(results)} results (Y/N)?")
        if resp.lower() == "n":
            print_stderr("\nExiting...", params)
            sys.exit(0)

        # Extra cosmetic newline
        print_stderr("", params)


def print_results_table(with_index, with_score, results, params):
    """Prepare and print the table of suggest search results."""
    # Field widths in output
    score_width = 7
    index_width = 7

    # This is necessary because the results are returned as a
    # list[str] if neither index nor score is included; but are
    # returned as a list[tuple[...]] if one or both of index/score
    # is included.
    if with_index or with_score:
        rst_width = max(len(res[0]) for res in results)
    else:
        rst_width = max(len(res) for res in results)

    rst_width += 2

    # For now, in each case the formatting for each row is dynamically
    # stored in `fmt`, and then `fmt` is used to actually format each row.

    # TODO: Consider replacing this with a *real* table formatting tool
    if with_index:
        if with_score:
            gen = generate_score_index_lines(
                results, score_width, index_width, rst_width
            )
        else:
            gen = generate_index_lines(results, index_width, rst_width)
    else:
        if with_score:
            gen = generate_score_lines(results, score_width, rst_width)
        else:
            gen = generate_names_only_lines(results)

    if not params[PrsConst.PAGINATE]:
        print("\n".join(gen))
    else:
        # To make sure the initial output is not scrolled off the screen
        # when --all is specified.
        if params[PrsConst.ALL]:
            input("Press Enter to continue...")

        while True:
            # Adjust the number of lines per page if the user changes their
            # terminal window size mid-execution
            n_lines = shutil.get_terminal_size().lines - 2
            out_text = "\n".join(itt.islice(gen, n_lines))

            if out_text:
                print(out_text)
                # Don't paginate after the last, partial screenful of output
                if out_text.count("\n") == n_lines - 1:
                    input("Press Enter to continue...")
            else:
                # join() above will supply an empty string once gen is exhausted
                break


def generate_score_index_lines(results, score_width, index_width, rst_width):
    """Yield lines to print the table with scores & indices."""
    fmt = (
        f"{{name: <{rst_width}}}  {{score: ^{score_width}}}  {{index: ^{index_width}}}"
    )
    yield fmt.format(name="  Name", score="Score", index="Index")
    yield fmt.format(
        name=("-" * rst_width), score=("-" * score_width), index=("-" * index_width)
    )
    yield from (fmt.format(name=res[0], score=res[1], index=res[2]) for res in results)


def generate_index_lines(results, index_width, rst_width):
    """Yield lines to print the table with indices."""
    fmt = f"{{name: <{rst_width}}}  {{index: ^{index_width}}}"
    yield fmt.format(name="  Name", index="Index")
    yield fmt.format(name=("-" * rst_width), index=("-" * index_width))
    yield from (fmt.format(name=res[0], index=res[1]) for res in results)


def generate_score_lines(results, score_width, rst_width):
    """Yield lines to print the table with scores."""
    fmt = f"{{name: <{rst_width}}}  {{score: ^{score_width}}}"
    yield fmt.format(name="  Name", score="Score")
    yield fmt.format(name=("-" * rst_width), score=("-" * score_width))
    yield from (fmt.format(name=res[0], score=res[1]) for res in results)


def generate_names_only_lines(results):
    """Yield lines to print containing just the object search results."""
    yield from (str(res) for res in results)


def print_stderr_inferred_mapping(params):
    """Print as good of an `intersphinx_mapping` entry as can be determined."""
    if not params[PrsConst.URL]:
        print_stderr(
            "Cannot infer intersphinx_mapping from a local objects.inv.\n", params
        )
        return

    input_url = params[PrsConst.INFILE]
    inv_url = params[PrsConst.FOUND_URL]

    reduced_inv_url = extract_objectsinv_url_base(inv_url)

    if input_url == inv_url:
        # User provided an exact URL to an inventory
        # The tail of input_url thus should not match anything in the inventory,
        # so we can't say anything about what the URL base of the docset is.
        if inv_url == reduced_inv_url:
            # The inventory URL does not end with the standard /objects.inv,
            # so we don't even have *that* point of reference to work with.
            print_stderr(
                (
                    "Cannot infer intersphinx_mapping for this docset using "
                    "the provided input URL.\n"
                ),
                params,
            )
        else:
            # The inventory URL *does* end with standard /objects.inv,
            # so there's a reasonable chance we know the mapping.
            print_stderr(
                (
                    "The intersphinx_mapping for this docset is PROBABLY:\n\n"
                    f"  ({reduced_inv_url}, None)\n"
                ),
                params,
            )
    else:
        # User provided a URL from the docset that reduced to a base, atop which
        # an inventory at .../objects.inv was found.
        if inv_url == reduced_inv_url:
            # This should never happen, because the only way we should be in
            # this outer else is if inv_url *DOES* end in /objects.inv
            print_stderr(  # pragma: no cover
                (
                    "ERROR: Inconsistent internal state "
                    "during intersphinx_mapping inference.\n"
                ),
                params,
            )
        else:
            # Here we're *very* confident that we've go the mapping.
            print_stderr(
                (
                    "The intersphinx_mapping for this docset is LIKELY:\n\n"
                    f"  ({reduced_inv_url}, None)\n"
                ),
                params,
            )


def extract_objectsinv_url_base(objectsinv_url):
    """Infer a base URL for the provided ``objects.inv`` inventory URL.

    If this function is a no-op, then the resulting base is NOT RELIABLE,
    because the URL did not end with ``/objects.inv``.

    If this function *does* make a change, then the resulting base is
    RELATIVELY RELIABLE, since the only change that should occur is
    stripping of a ``/objects.inv`` suffix, which strongly implies but
    does not guarantee that the URL came from a Sphinx docset in the
    standard multi-page HTML layout.

    Parameters
    ----------
    objectsinv_url

        |str| -- URL from which to attempt docset base inference

    Returns
    -------
    trimmed

        |str| -- URL after attempt to trim a trailing ``/objects.inv``

    """
    trimmed = _strip_url_to_netloc_path(objectsinv_url, with_scheme=True)
    base = trimmed.rpartition("/objects.inv")[0]

    if base:
        return f"{base}/"
    else:
        return objectsinv_url


def _strip_url_to_netloc_path(url, *, with_scheme=False):
    """Reduce a URL to only netloc and path, optionally with scheme."""
    parts = urlparse.urlsplit(url)
    trimmed = parts._replace(
        query="",
        fragment="",
    )

    if not with_scheme:
        trimmed = trimmed._replace(scheme="")

    return urlparse.urlunsplit(trimmed)



r"""``sphobjinv`` *CLI parser definition module*.

``sphobjinv`` is a toolkit for manipulation and inspection of
Sphinx |objects.inv| files.

**Author**
    Brian Skinn (brian.skinn@gmail.com)

**File Created**
    15 Nov 2020

**Copyright**
    \(c) 2016-2026 Brian Skinn and community contributors

**Source Repository**
    https://github.com/bskinn/sphobjinv

**Documentation**
    https://sphobjinv.readthedocs.io/en/stable

**License**
    Code: `MIT License`_

    Docs & Docstrings: |CC BY 4.0|_

    See |license_txt|_ for full license terms.

**Members**

"""

import argparse as ap

from sphobjinv.version import __version__


class PrsConst:
    """Container for CLI parser constants."""

    # ### Version arg and helpers
    #: Optional argument name for use with the base
    #: argument parser, to show version &c. info, and exit
    VERSION = "version"

    #: Version &c. output blurb
    VER_TXT = (
        f"\nsphobjinv v{__version__}\n\n"
        "Copyright (c) 2016-2026 Brian Skinn and community contributors\n"
        "License: The MIT License\n\n"
        "Bug reports & feature requests:"
        " https://github.com/bskinn/sphobjinv\n"
        "Documentation:"
        " https://sphobjinv.readthedocs.io\n"
    )

    #: Short version text for textconv entrypoint
    VER_TXT_SHORT = f"sphobjinv v{__version__}"

    # ### Subparser selectors and argparse param for storing subparser name
    #: Subparser name for inventory file conversions; stored in
    #: :data:`SUBPARSER_NAME` when selected
    CONVERT = "convert"

    #: Subparser name for inventory object suggestions; stored in
    #: :data:`SUBPARSER_NAME` when selected
    SUGGEST = "suggest"

    #: Param for storing subparser name
    #: (:data:`CONVERT` or :data:`SUGGEST`)
    SUBPARSER_NAME = "sprs_name"

    # ### Common URL argument for both subparsers
    #: Optional argument name for use with both :data:`CONVERT` and
    #: :data:`SUGGEST` subparsers, indicating that
    #: :data:`INFILE` is to be treated as a URL
    #: rather than a local file path
    URL = "url"

    # ### Conversion subparser: 'mode' param and choices
    #: Positional argument name for use with :data:`CONVERT` subparser,
    #: indicating output file format
    #: (:data:`ZLIB`, :data:`PLAIN` or :data:`JSON`)
    MODE = "mode"

    #: Argument value for :data:`CONVERT` :data:`MODE`,
    #: to output a :mod:`zlib`-compressed inventory
    ZLIB = "zlib"

    #: Argument value for :data:`CONVERT` :data:`MODE`,
    #: to output a plaintext inventory
    PLAIN = "plain"

    #: Argument value for :data:`CONVERT` :data:`MODE`,
    #: to output an inventory as JSON
    JSON = "json"

    # ### Source/destination params
    #: Required positional argument name for use with both :data:`CONVERT` and
    #: :data:`SUGGEST` subparsers, holding the path
    #: (or URL, if :data:`URL` is specified)
    #: to the input file
    INFILE = "infile"

    #: Optional positional argument name
    #: for use with the :data:`CONVERT` subparser,
    #: holding the path to the output file
    #: (:data:`DEF_BASENAME` and the appropriate item from :data:`DEF_OUT_EXT`
    #: are used if this argument is not provided)
    OUTFILE = "outfile"

    # ### Convert subparser optional params
    #: Optional argument name for use with the :data:`CONVERT` subparser,
    #: indicating to suppress console output
    QUIET = "quiet"

    #: Optional argument name for use with the :data:`CONVERT` subparser,
    #: indicating to expand URI and display name
    #: abbreviations in the generated output file
    EXPAND = "expand"

    #: Optional argument name for use with the :data:`CONVERT` subparser,
    #: indicating to contract URIs and display names
    #: to abbreviated forms in the generated output file
    CONTRACT = "contract"

    #: Optional argument name for use with the :data:`CONVERT` subparser,
    #: indicating to overwrite any existing output
    #: file without prompting
    OVERWRITE = "overwrite"

    # ### Suggest subparser params
    #: Positional argument name for use with the :data:`SUGGEST` subparser,
    #: holding the search term for |fuzzywuzzy|_ text matching
    SEARCH = "search"

    #: Optional argument name for use with the :data:`SUGGEST` subparser,
    #: taking the minimum desired |fuzzywuzzy|_ match quality
    #: as one required argument
    THRESH = "thresh"

    #: Optional argument name for use with the :data:`SUGGEST` subparser,
    #: indicating to print the location index of each returned object
    #: within :data:`INFILE` along with the object domain/role/name
    #: (may be specified with :data:`SCORE`)
    INDEX = "index"

    #: Optional argument name for use with the :data:`SUGGEST` subparser,
    #: indicating to print the |fuzzywuzzy|_ score of each returned object
    #: within :data:`INFILE` along with the object domain/role/name
    #: (may be specified with :data:`INDEX`)
    SCORE = "score"

    #: Optional argument name for use with the :data:`SUGGEST` subparser,
    #: indicating to print all returned objects, regardless of the
    #: number returned, without asking for confirmation
    ALL = "all"

    #: Optional argument name for use with the :data:`SUGGEST` subparser,
    #: indicating to paginate the suggest subcommand results
    PAGINATE = "paginate"

    # ### Helper strings
    #: Help text for the :data:`CONVERT` subparser
    HELP_CO_PARSER = (
        "Convert intersphinx inventory to zlib-compressed, plaintext, or JSON formats."
    )

    #: Help text for the :data:`SUGGEST` subparser
    HELP_SU_PARSER = "Fuzzy-search intersphinx inventory for desired object(s)."

    #: Help text for default extensions for the various conversion types
    HELP_CONV_EXTS = "'.inv/.txt/.json'"

    # ### Defaults for an unspecified OUTFILE
    #: Default base name for an unspecified :data:`OUTFILE`
    DEF_BASENAME = "objects"

    #: Default extensions for an unspecified :data:`OUTFILE`
    DEF_OUT_EXT = {ZLIB: ".inv", PLAIN: ".txt", JSON: ".json"}

    # ### Useful constants
    #: Number of returned objects from a :data:`SUGGEST` subparser invocation
    #: above which user will be prompted for confirmation to print the results
    #: (unless :data:`ALL` is specified)
    SUGGEST_CONFIRM_LENGTH = 30

    #: Default match threshold for :option:`sphobjinv suggest --thresh`
    DEF_THRESH = 75

    #: Dict key for URL at which an inventory was actually found
    FOUND_URL = "found_url"


def getparser():
    """Generate argument parser.

    Returns
    -------
    prs

        :class:`~argparse.ArgumentParser` -- Parser for commandline usage
        of |soi|

    """
    prs = ap.ArgumentParser(
        description="Format conversion for "
        "and introspection of "
        "intersphinx "
        "'objects.inv' files."
    )
    prs.add_argument(
        "-" + PrsConst.VERSION[0],
        "--" + PrsConst.VERSION,
        help="Print package version & other info",
        action="store_true",
    )

    sprs = prs.add_subparsers(
        title="Subcommands",
        dest=PrsConst.SUBPARSER_NAME,
        metavar=f"{{{PrsConst.CONVERT},{PrsConst.SUGGEST}}}",
        help="Execution mode. Type "
        "'sphobjinv [mode] -h' "
        "for more information "
        "on available options. "
        "Mode names can be abbreviated "
        "to their first two letters.",
    )

    # Enforce subparser as optional. No effect for 3.4 to 3.7;
    # briefly required a/o 3.7.0b4 due to change in default behavior, per:
    # https://bugs.python.org/issue33109. 3.6 behavior restored for
    # 3.7 release.
    #
    # We retain for explicitness
    sprs.required = False

    spr_convert = sprs.add_parser(
        PrsConst.CONVERT,
        aliases=[PrsConst.CONVERT[:2]],
        help=PrsConst.HELP_CO_PARSER,
        description=PrsConst.HELP_CO_PARSER,
    )
    spr_suggest = sprs.add_parser(
        PrsConst.SUGGEST,
        aliases=[PrsConst.SUGGEST[:2]],
        help=PrsConst.HELP_SU_PARSER,
        description=PrsConst.HELP_SU_PARSER,
    )

    # ### Args for conversion subparser
    spr_convert.add_argument(
        PrsConst.MODE,
        help="Conversion output format",
        choices=(PrsConst.ZLIB, PrsConst.PLAIN, PrsConst.JSON),
    )

    spr_convert.add_argument(
        PrsConst.INFILE,
        help=(
            "Path to file to be converted. Passing '-' indicates to read from stdin "
            "(plaintext/JSON only)."
        ),
    )

    spr_convert.add_argument(
        PrsConst.OUTFILE,
        help=(
            "Path to desired output file. "
            "Defaults to same directory and main "
            "file name as input file but with extension "
            + PrsConst.HELP_CONV_EXTS
            + ", as appropriate for the output format. "
            "A path to a directory is accepted here, "
            "in which case the default output file name will be used. "
            "Passing '-' indicates to write to stdout. If "
            + PrsConst.INFILE
            + " is passed as '-', "
            + PrsConst.OUTFILE
            + " can be omitted and both stdin and stdout will be used."
        ),
        nargs="?",
        default=None,
    )

    # Mutually exclusive group for --expand/--contract
    gp_expcont = spr_convert.add_argument_group(title="URI/display name conversions")
    meg_expcont = gp_expcont.add_mutually_exclusive_group()
    meg_expcont.add_argument(
        "-" + PrsConst.EXPAND[0],
        "--" + PrsConst.EXPAND,
        help="Expand all URI and display name abbreviations",
        action="store_true",
    )

    meg_expcont.add_argument(
        "-" + PrsConst.CONTRACT[0],
        "--" + PrsConst.CONTRACT,
        help="Contract all URI and display name abbreviations",
        action="store_true",
    )

    # Clobber argument
    spr_convert.add_argument(
        "-" + PrsConst.OVERWRITE[0],
        "--" + PrsConst.OVERWRITE,
        help="Overwrite output files without prompting",
        action="store_true",
    )

    # stdout suppressor option (e.g., for scripting)
    spr_convert.add_argument(
        "-" + PrsConst.QUIET[0],
        "--" + PrsConst.QUIET,
        help="Suppress printing of status messages and "
        "overwrite output files without prompting",
        action="store_true",
    )

    # Flag to treat infile as a URL
    spr_convert.add_argument(
        "-" + PrsConst.URL[0],
        "--" + PrsConst.URL,
        help=(
            "Treat 'infile' as a URL for download. "
            "Cannot be used with an infile of '-'."
        ),
        action="store_true",
    )

    # ### Args for suggest subparser
    spr_suggest.add_argument(
        PrsConst.INFILE,
        help=(
            "Path to inventory file to be searched. "
            "Passing '-' indicates to read from stdin (plaintext/JSON only)."
        ),
    )
    spr_suggest.add_argument(PrsConst.SEARCH, help="Search term for object suggestions")
    spr_suggest.add_argument(
        "-" + PrsConst.ALL[0],
        "--" + PrsConst.ALL,
        help="Display all results "
        "regardless of the number returned "
        "without prompting for confirmation.",
        action="store_true",
    )
    spr_suggest.add_argument(
        "-" + PrsConst.PAGINATE[0],
        "--" + PrsConst.PAGINATE,
        help="Paginate long search results",
        action="store_true",
    )
    spr_suggest.add_argument(
        "-" + PrsConst.INDEX[0],
        "--" + PrsConst.INDEX,
        help="Include Inventory.objects list indices with the search results",
        action="store_true",
    )
    spr_suggest.add_argument(
        "-" + PrsConst.SCORE[0],
        "--" + PrsConst.SCORE,
        help="Include fuzzywuzzy scores with the search results",
        action="store_true",
    )
    spr_suggest.add_argument(
        "-" + PrsConst.THRESH[0],
        "--" + PrsConst.THRESH,
        help="Match quality threshold, integer 0-100, "
        "default 75. Default is suitable when "
        "'search' is exactly a known object name. "
        "A value of 30-50 gives better results "
        "for approximate matches.",
        default=PrsConst.DEF_THRESH,
        type=int,
        choices=range(101),
        metavar="{0-100}",
    )
    spr_suggest.add_argument(
        "-" + PrsConst.URL[0],
        "--" + PrsConst.URL,
        help=(
            "Treat 'infile' as a URL for download. "
            f"Cannot be used with --{PrsConst.URL}."
        ),
        action="store_true",
    )

    return prs


def getparser_textconv():
    """Generate argument parser for textconv entrypoint.

    Returns
    -------
    prs

        :class:`~argparse.ArgumentParser` -- Parser for textconv commandline
        usage of |soi|

    """
    prs = ap.ArgumentParser(
        description=(
            "Emit the plaintext of the local Sphinx inventory at 'infile' to stdout."
        )
    )
    prs.add_argument(
        "-" + PrsConst.VERSION[0],
        "--" + PrsConst.VERSION,
        help="Print package version & other info",
        action="version",
        version=PrsConst.VER_TXT_SHORT,
    )

    prs.add_argument(
        PrsConst.INFILE,
        help=("Path to file to be converted"),
    )

    return prs



r"""*Helper enums for* ``sphobjinv``.

``sphobjinv`` is a toolkit for manipulation and inspection of
Sphinx |objects.inv| files.

**Author**
    Brian Skinn (brian.skinn@gmail.com)

**File Created**
    4 May 2019

**Copyright**
    \(c) 2016-2026 Brian Skinn and community contributors

**Source Repository**
    https://github.com/bskinn/sphobjinv

**Documentation**
    https://sphobjinv.readthedocs.io/en/stable

**License**
    Code: `MIT License`_

    Docs & Docstrings: |CC BY 4.0|_

    See |license_txt|_ for full license terms.

**Members**

"""

from enum import Enum


class SourceTypes(Enum):
    """|Enum| for the import mode used in instantiating an |Inventory|.

    Since |Enum| keys iterate in definition order, the
    definition order here defines the order in which |Inventory|
    objects attempt to parse a source object passed to
    :class:`Inventory.__init__() <sphobjinv.inventory.Inventory>`
    either as a positional argument
    or via the generic `source` keyword argument.

    This order **DIFFERS** from the documentation order, which is
    alphabetical.

    """

    #: No source; |Inventory| was instantiated with
    #: :data:`~sphobjinv.inventory.Inventory.project`
    #: and :data:`~sphobjinv.inventory.Inventory.version`
    #: as empty strings and
    #: :data:`~sphobjinv.inventory.Inventory.objects` as an empty |list|.
    Manual = "manual"

    #: Instantiation from a plaintext |objects.inv| |bytes|.
    BytesPlaintext = "bytes_plain"

    #: Instantiation from a zlib-compressed
    #: |objects.inv| |bytes|.
    BytesZlib = "bytes_zlib"

    #: Instantiation from a plaintext |objects.inv| file on disk.
    FnamePlaintext = "fname_plain"

    #: Instantiation from a zlib-compressed |objects.inv| file on disk.
    FnameZlib = "fname_zlib"

    #: Instantiation from a |dict| validated against
    #: :data:`schema.json_schema <sphobjinv.schema.json_schema>`.
    DictJSON = "dict_json"

    #: Instantiation from a zlib-compressed |objects.inv| file
    #: downloaded from a URL.
    URL = "url"


class HeaderFields(Enum):
    """|Enum| for various inventory-level data items.

    A subset of these |Enum| values is used in various Regex,
    JSON, and string formatting contexts within
    class:`~sphobjinv.inventory.Inventory`
    and :data:`schema.json_schema <sphobjinv.schema.json_schema>`.

    """

    #: Project name associated with an inventory
    Project = "project"

    #: Project version associated with an inventory
    Version = "version"

    #: Number of objects contained in the inventory
    Count = "count"

    #: The |str| value of this |Enum| member is accepted as a root-level
    #: key in a |dict| to be imported into an
    #: :class:`~sphobjinv.inventory.Inventory`.
    #: The corresponding value in the |dict| may contain any arbitrary data.
    #: Its possible presence is accounted for in
    #: :data:`schema.json_schema <sphobjinv.schema.json_schema>`.
    #:
    #: The data associated with this key are **ignored**
    #: during import into an
    #: :class:`~sphobjinv.inventory.Inventory`.
    Metadata = "metadata"



r"""``sphobjinv`` *CLI core execution module*.

``sphobjinv`` is a toolkit for manipulation and inspection of
Sphinx |objects.inv| files.

**Author**
    Brian Skinn (brian.skinn@gmail.com)

**File Created**
    15 Nov 2020

**Copyright**
    \(c) 2016-2026 Brian Skinn and community contributors

**Source Repository**
    https://github.com/bskinn/sphobjinv

**Documentation**
    https://sphobjinv.readthedocs.io/en/stable

**License**
    Code: `MIT License`_

    Docs & Docstrings: |CC BY 4.0|_

    See |license_txt|_ for full license terms.

**Members**

"""

import sys

from sphobjinv.cli.convert import do_convert
from sphobjinv.cli.load import inv_local, inv_stdin, inv_url
from sphobjinv.cli.parser import PrsConst, getparser, getparser_textconv
from sphobjinv.cli.suggest import do_suggest
from sphobjinv.cli.ui import print_stderr


def main():
    r"""Handle command line invocation.

    Parses command line arguments,
    handling the no-arguments and
    VERSION cases.

    Creates the |Inventory| from the indicated source
    and method.

    Invokes :func:`~sphobjinv.cli.convert.do_convert` or
    :func:`~sphobjinv.cli.suggest.do_suggest`
    per the subparser name stored in SUBPARSER_NAME.

    """
    # If no args passed, stick in '-h'
    if len(sys.argv) == 1:
        sys.argv.append("-h")

    # Parse commandline arguments, discarding any unknown ones
    # I forget why I set it up to discard these, it might be
    # more confusing than it's worth to swallow them this way....
    prs = getparser()
    ns, _ = prs.parse_known_args()
    params = vars(ns)

    # Print version &c. and exit if indicated
    if params[PrsConst.VERSION]:
        print(PrsConst.VER_TXT)
        sys.exit(0)

    # At this point, need to trap for a null subparser
    if not params[PrsConst.SUBPARSER_NAME]:
        prs.error("No subparser selected")

    # Regardless of mode, insert extra blank line
    # for cosmetics
    print_stderr(" ", params)

    # Generate the input Inventory based on --url or stdio or file.
    # These inventory-load functions should call
    # sys.exit(n) internally in error-exit situations
    if params[PrsConst.URL]:
        if params[PrsConst.INFILE] == "-":
            prs.error("argument -u/--url not allowed with '-' as infile")
        inv, in_path = inv_url(params)
    elif params[PrsConst.INFILE] == "-":
        inv = inv_stdin(params)
        in_path = None
    else:
        inv, in_path = inv_local(params)

    # Perform action based upon mode
    if params[PrsConst.SUBPARSER_NAME][:2] == PrsConst.CONVERT[:2]:
        do_convert(inv, in_path, params)
    elif params[PrsConst.SUBPARSER_NAME][:2] == PrsConst.SUGGEST[:2]:
        do_suggest(inv, params)

    # Cosmetic final blank line
    print_stderr(" ", params)

    # Clean exit
    sys.exit(0)


def main_textconv():
    """Entrypoint for textconv operation."""
    # If no args passed, stick in '-h'
    if len(sys.argv) == 1:
        sys.argv.append("-h")

    prs = getparser_textconv()
    params = vars(prs.parse_args())

    # No version arg handling, using 'version' action in this parser

    inv, in_path = inv_local(params)

    params[PrsConst.CONTRACT] = False
    params[PrsConst.EXPAND] = False
    params[PrsConst.MODE] = PrsConst.PLAIN
    params[PrsConst.OUTFILE] = "-"

    do_convert(inv, in_path, params)

    sys.exit(0)



#!/usr/bin/env python
# encoding: utf-8
"""
score.py

Copyright (c) 2011 Adam Cohen

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import sys
import os
import re
from difflib import SequenceMatcher
from sphobjinv._vendored.fuzzywuzzy import utils

REG_TOKEN = re.compile(r"[\w\d]+")  # B Skinn 2021-12-11

###########################
# Basic Scoring Functions #
###########################

def ratio(s1,  s2):

    if s1 is None: raise TypeError("s1 is None")
    if s2 is None: raise TypeError("s2 is None")

    m = SequenceMatcher(None, s1, s2)
    return int(100 * m.ratio())

# todo: skip duplicate indexes for a little more speed
def partial_ratio(s1,  s2):

    if s1 is None: raise TypeError("s1 is None")
    if s2 is None: raise TypeError("s2 is None")

    if len(s1) <= len(s2):
        shorter = s1; longer = s2;
    else:
        shorter = s2; longer = s1

    m = SequenceMatcher(None, shorter, longer)
    blocks = m.get_matching_blocks()

    # each block represents a sequence of matching characters in a string
    # of the form (idx_1, idx_2, len)
    # the best partial match will block align with at least one of those blocks
    #   e.g. shorter = "abcd", longer = XXXbcdeEEE
    #   block = (1,3,3)
    #   best score === ratio("abcd", "Xbcd")
    scores = []
    for block in blocks:
        long_start   = block[1] - block[0] if (block[1] - block[0]) > 0 else 0
        long_end     = long_start + len(shorter)
        long_substr  = longer[long_start:long_end]

        m2 = SequenceMatcher(None, shorter, long_substr)
        r = m2.ratio()
        if r > .995: return 100
        else: scores.append(r)

    return int(100 * max(scores))

##############################
# Advanced Scoring Functions #
##############################

# Sorted Token
#   find all alphanumeric tokens in the string
#   sort those tokens and take ratio of resulting joined strings
#   controls for unordered string elements
def _token_sort(s1,  s2, partial=True):

    if s1 is None: raise TypeError("s1 is None")
    if s2 is None: raise TypeError("s2 is None")

    # pull tokens
    tokens1 = REG_TOKEN.findall(s1)
    tokens2 = REG_TOKEN.findall(s2)

    # sort tokens and join
    sorted1 = u" ".join(sorted(tokens1))
    sorted2 = u" ".join(sorted(tokens2))

    sorted1 = sorted1.strip()
    sorted2 = sorted2.strip()

    if partial:
        return partial_ratio(sorted1, sorted2)
    else:
        return ratio(sorted1, sorted2)

def token_sort_ratio(s1,  s2):
    return _token_sort(s1, s2, False)

def partial_token_sort_ratio(s1,  s2):
    return _token_sort(s1, s2, True)

# Token Set
#   find all alphanumeric tokens in each string...treat them as a set
#   construct two strings of the form
#       <sorted_intersection><sorted_remainder>
#   take ratios of those two strings
#   controls for unordered partial matches
def _token_set(s1,  s2, partial=True):

    if s1 is None: raise TypeError("s1 is None")
    if s2 is None: raise TypeError("s2 is None")

    # pull tokens
    tokens1 = set(REG_TOKEN.findall(s1))
    tokens2 = set(REG_TOKEN.findall(s2))

    intersection = tokens1.intersection(tokens2)
    diff1to2 = tokens1.difference(tokens2)
    diff2to1 = tokens2.difference(tokens1)

    sorted_sect = u" ".join(sorted(intersection))
    sorted_1to2 = u" ".join(sorted(diff1to2))
    sorted_2to1 = u" ".join(sorted(diff2to1))

    combined_1to2 = sorted_sect + " " + sorted_1to2
    combined_2to1 = sorted_sect + " " + sorted_2to1

    # strip
    sorted_sect = sorted_sect.strip()
    combined_1to2 = combined_1to2.strip()
    combined_2to1 = combined_2to1.strip()

    pairwise = [
        ratio(sorted_sect, combined_1to2),
        ratio(sorted_sect, combined_2to1),
        ratio(combined_1to2, combined_2to1)
    ]
    return max(pairwise)

    # if partial:
    #     # partial_token_set_ratio
    #
    # else:
    #     # token_set_ratio
    #     tsr = ratio(combined_1to2, combined_2to1)
    #     return tsr

def token_set_ratio(s1,  s2):
    return _token_set(s1, s2, False)

def partial_token_set_ratio(s1,  s2):
    return _token_set(s1, s2, True)

# TODO: numerics

###################
# Combination API #
###################

# q is for quick
def QRatio(s1,  s2):
    if not utils.validate_string(s1): return 0
    if not utils.validate_string(s2): return 0

    p1 = utils.full_process(s1)
    p2 = utils.full_process(s2)

    return ratio(p1, p2)

# w is for weighted
def WRatio(s1,  s2):
    p1 = utils.full_process(s1)
    p2 = utils.full_process(s2)
    if not utils.validate_string(p1): return 0
    if not utils.validate_string(p2): return 0

    # should we look at partials?
    try_partial     = True
    unbase_scale    = .95
    partial_scale   = .90

    base = ratio(p1, p2)
    len_ratio = float(max(len(p1),len(p2)))/min(len(p1),len(p2))

    # if strings are similar length, don't use partials
    if len_ratio < 1.5: try_partial = False

    # if one string is much much shorter than the other
    if len_ratio > 8: partial_scale = .6

    if try_partial:
        partial      = partial_ratio(p1, p2)                 * partial_scale
        ptsor        = partial_token_sort_ratio(p1, p2)      * unbase_scale * partial_scale
        ptser        = partial_token_set_ratio(p1, p2)       * unbase_scale * partial_scale

        return int(max(base, partial, ptsor, ptser))
    else:
        tsor         = token_sort_ratio(p1, p2)              * unbase_scale
        tser         = token_set_ratio(p1, p2)               * unbase_scale

        return int(max(base, tsor, tser))




r"""*Module for* ``sphobjinv`` *CLI* |Inventory| *loading*.

``sphobjinv`` is a toolkit for manipulation and inspection of
Sphinx |objects.inv| files.

**Author**
    Brian Skinn (brian.skinn@gmail.com)

**File Created**
    17 Nov 2020

**Copyright**
    \(c) 2016-2026 Brian Skinn and community contributors

**Source Repository**
    https://github.com/bskinn/sphobjinv

**Documentation**
    https://sphobjinv.readthedocs.io/en/stable

**License**
    Code: `MIT License`_

    Docs & Docstrings: |CC BY 4.0|_

    See |license_txt|_ for full license terms.

**Members**

"""

import json
import sys
from json import JSONDecodeError
from urllib.error import HTTPError, URLError

from jsonschema.exceptions import ValidationError

from sphobjinv import Inventory, VersionError, readjson, urlwalk
from sphobjinv.cli.parser import PrsConst
from sphobjinv.cli.paths import resolve_inpath
from sphobjinv.cli.ui import err_format, print_stderr


def import_infile(in_path):
    """Attempt import of indicated file.

    Convenience function wrapping attempts to load an
    |Inventory| from a local path.

    Parameters
    ----------
    in_path

        |str| -- Path to input file

    Returns
    -------
    inv

        |Inventory| or |None| -- If instantiation with the file at
        `in_path` succeeds, the resulting |Inventory| instance;
        otherwise, |None|

    """
    # Try general import, for zlib or plaintext files
    try:
        inv = Inventory(in_path)
    except AttributeError:
        pass  # Punt to JSON attempt
    else:
        return inv

    # Maybe it's JSON
    try:
        inv = Inventory(readjson(in_path))
    except JSONDecodeError:
        return None
    else:
        return inv


def inv_local(params):
    """Create |Inventory| from local source.

    Uses ``resolve_inpath`` to sanity-check and/or convert
    INFILE.

    Calls :func:`sys.exit` internally in error-exit situations.

    Parameters
    ----------
    params

        |dict| -- Parameters/values mapping from the active subparser

    Returns
    -------
    inv

        |Inventory| -- Object representation of the inventory
        at INFILE

    in_path

        |str| -- Input file path as resolved/checked by
        ``resolve_inpath``

    """
    # Resolve input file path
    try:
        in_path = resolve_inpath(params[PrsConst.INFILE])
    except Exception as e:
        print_stderr("\nError while parsing input file path:", params)
        print_stderr(err_format(e), params)
        sys.exit(1)

    # Attempt import
    inv = import_infile(in_path)
    if inv is None:
        print_stderr("\nError: Unrecognized file format", params)
        sys.exit(1)

    return inv, in_path


def inv_url(params):
    """Create |Inventory| from file downloaded from URL.

    Initially, treats INFILE as a download URL to be passed to
    the `url` initialization argument
    of :class:`~sphobjinv.inventory.Inventory`.

    If an inventory is not found at that exact URL, progressively
    searches the directory tree of the URL for |objects.inv|.

    Injects the URL at which an inventory was found into `params`
    under the FOUND_URL key.

    Calls :func:`sys.exit` internally in error-exit situations.

    Parameters
    ----------
    params

        |dict| -- Parameters/values mapping from the active subparser

    Returns
    -------
    inv

        |Inventory| -- Object representation of the inventory
        at INFILE

    ret_path

        |str| -- URL from INFILE used to construct `inv`.
        If URL is longer than 45 characters, the central portion is elided.

    """
    in_file = params[PrsConst.INFILE]

    def attempt_inv_load(url, params):
        """Attempt the Inventory load and report outcome."""
        inv = None

        try:
            inv = Inventory(url=url)
        except HTTPError as e:
            print_stderr(f"  ... HTTP error: {e.code} {e.reason}.", params)
        except URLError:  # pragma: no cover
            print_stderr("  ... error attempting to retrieve URL.", params)
        except VersionError:  # pragma: no cover
            print_stderr("  ... no recognized inventory.", params)
        except ValueError:
            print_stderr(
                (
                    "  ... file found but inventory could not be loaded. "
                    "(Did you forget https:// ?)"
                ),
                params,
            )
        else:
            print_stderr("  ... inventory found.", params)

        return inv

    # Disallow --url mode on local files
    if in_file.startswith("file:/"):
        print_stderr("\nError: URL mode on local file is invalid", params)
        sys.exit(1)

    print_stderr(f"Attempting {in_file} ...", params)
    inv = attempt_inv_load(in_file, params)

    if inv:
        url = in_file
    else:
        for url in urlwalk(in_file):
            print_stderr(f'Attempting "{url}" ...', params)
            inv = attempt_inv_load(url, params)
            if inv:
                break

    # Cosmetic line break
    print_stderr(" ", params)

    # Success or no?
    if not inv:
        print_stderr("No inventory found!", params)
        sys.exit(1)

    params.update({PrsConst.FOUND_URL: url})
    if len(url) > 45:
        ret_path = url[:20] + "[...]" + url[-20:]
    else:  # pragma: no cover
        ret_path = url

    return inv, ret_path


def inv_stdin(params):
    """Create |Inventory| from contents of stdin.

    Due to stdin's encoding and formatting assumptions, only
    text-based inventory formats can be sanely parsed.

    Thus, only plaintext and JSON inventory formats can be
    used as inputs here.

    Parameters
    ----------
    params

        |dict| -- Parameters/values mapping from the active subparser

    Returns
    -------
    inv

        |Inventory| -- Object representation of the inventory
        provided at stdin

    """
    data = sys.stdin.read()

    try:
        return Inventory(dict_json=json.loads(data))
    except (JSONDecodeError, ValidationError):
        pass

    try:
        return Inventory(plaintext=data)
    except (AttributeError, UnicodeEncodeError, TypeError):
        pass

    print_stderr("Invalid plaintext or JSON inventory format.", params)
    sys.exit(1)



r"""``sphobjinv`` *data classes for individual objects*.

``sphobjinv`` is a toolkit for manipulation and inspection of
Sphinx |objects.inv| files.

**Author**
    Brian Skinn (brian.skinn@gmail.com)

**File Created**
    7 Nov 2017

**Copyright**
    \(c) 2016-2026 Brian Skinn and community contributors

**Source Repository**
    https://github.com/bskinn/sphobjinv

**Documentation**
    https://sphobjinv.readthedocs.io/en/stable

**License**
    Code: `MIT License`_

    Docs & Docstrings: |CC BY 4.0|_

    See |license_txt|_ for full license terms.

**Members**

"""

from abc import ABCMeta, abstractmethod
from enum import Enum

import attr


class DataFields(Enum):
    """|Enum| for the fields of |objects.inv| data objects."""

    #: Object name, as recognized internally by Sphinx
    Name = "name"

    #: Sphinx domain housing the object
    Domain = "domain"

    #: Full name of Sphinx role to be used when referencing the object
    Role = "role"

    #: Object search priority
    Priority = "priority"

    #: URI to the location of the object's documentation,
    #: relative to the documentation root
    URI = "uri"

    #: Default display name for the object
    #: in rendered documentation
    #: when referenced as
    #: |cour|\ \:domain\:role\:\`name\`\ |/cour|
    DispName = "dispname"


def _utf8_decode(b):
    """Decode (if needed) to str.

    Helper for type conversions among DataObjStr and DataObjBytes.

    """
    if isinstance(b, bytes):
        return b.decode(encoding="utf-8")
    elif isinstance(b, str):
        return b
    else:
        raise TypeError("Argument must be 'bytes' or 'str'")


def _utf8_encode(s):
    """Encode (if needed) to bytes.

    Helper for type conversions among DataObjStr and DataObjBytes.

    """
    if isinstance(s, str):
        return s.encode(encoding="utf-8")
    elif isinstance(s, bytes):
        return s
    else:
        raise TypeError("Argument must be 'bytes' or 'str'")


class SuperDataObj(metaclass=ABCMeta):
    """Abstract base superclass defining common methods &c. for data objects.

    Intended only to be subclassed
    by :class:`DataObjBytes` and :class:`DataObjStr`,
    to allow definition of common methods, properties, etc.
    all in one place.

    Where marked with |dag|,
    :class:`DataObjBytes` instances will return |bytes| values, whereas
    :class:`DataObjStr` instances will return |str| values.

    """

    #: Helper |str| for generating plaintext |objects.inv|
    #: data lines. The field names MUST match the |str| values
    #: of the :class:`~DataFields` members.
    data_line_fmt = "{name} {domain}:{role} {priority} {uri} {dispname}"

    #: |str.format| template for generating reST-like representations
    #: of object data for :data:`as_rst` (used with
    #: :meth:`Inventory.suggest() <sphobjinv.inventory.Inventory.suggest>`).
    rst_fmt = ":{domain}:{role}:`{name}`"

    def __str__(self):  # pragma: no cover
        """Return pretty string representation."""
        fmt_str = "<{0}:: :{1}:{2}:`{3}`>"

        return fmt_str.format(type(self).__name__, self.domain, self.role, self.name)

    @property
    @abstractmethod
    def name(self):
        r"""Object name, as recognized internally by Sphinx\ |dag|."""

    @property
    @abstractmethod
    def domain(self):
        r"""Sphinx domain containing the object\ |dag|."""

    @property
    @abstractmethod
    def role(self):
        r"""Sphinx role to be used when referencing the object\ |dag|."""

    @property
    @abstractmethod
    def priority(self):
        r"""Object search priority, as handled internally by Sphinx\ |dag|."""

    @property
    @abstractmethod
    def uri(self):
        r"""Object URI relative to documentation root\ |dag|.

        Possibly abbreviated; see :ref:`here <syntax_shorthand>`.

        """

    @property
    @abstractmethod
    def dispname(self):
        r"""Object default name in rendered documentation\ |dag|.

        Possibly abbreviated; see :ref:`here <syntax_shorthand>`.

        """

    @property
    @abstractmethod
    def uri_abbrev(self):
        r"""Abbreviation character(s) for URI tail\ |dag|.

        ``'$'`` or ``b'$'``
        for :doc:`version 2 </syntax>` |objects.inv| files.

        """

    @property
    @abstractmethod
    def dispname_abbrev(self):
        r"""Abbreviation character(s) for display name\ |dag|.

        ``'-'`` or ``b'-'``
        for :doc:`version 2 </syntax>` |objects.inv| files.

        """

    @property
    @abstractmethod
    def as_str(self, s):
        """:class:`DataObjStr` version of instance."""

    @property
    @abstractmethod
    def as_bytes(self, s):
        """:class:`DataObjBytes` version of instance."""

    @abstractmethod
    def _data_line_postprocess(self, s):
        """Post-process the data_line chars output."""

    @property
    def uri_contracted(self):
        """Object-relative URI, contracted with `uri_abbrev`."""
        if self.uri.endswith(self.name):
            return self.uri[: -len(self.name)] + self.uri_abbrev
        else:
            return self.uri

    @property
    def uri_expanded(self):
        """Object-relative URI, with `uri_abbrev` expanded."""
        if self.uri.endswith(self.uri_abbrev):
            return self.uri[: -len(self.uri_abbrev)] + self.name
        else:
            return self.uri

    @property
    def dispname_contracted(self):
        """Object display name, contracted with `dispname_abbrev`."""
        if self.dispname == self.name:
            return self.dispname_abbrev
        else:
            return self.dispname

    @property
    def dispname_expanded(self):
        """Object display name, with `dispname_abbrev` expanded."""
        if self.dispname == self.dispname_abbrev:
            return self.name
        else:
            return self.dispname

    @property
    def as_rst(self):
        r"""|str| reST reference-like object representation.

        Typically will NOT function as a proper reST reference
        in Sphinx source (e.g., a `role` of
        |cour|\ function\ |/cour| must be referenced using
        |cour|\ \:func\:\ |/cour| for the
        |cour|\ py\ |/cour| domain).
        """
        return self.rst_fmt.format(**self.as_str.json_dict())

    def json_dict(self, *, expand=False, contract=False):
        r"""Return the object data formatted as a flat |dict|.

        The returned |dict| is constructed such that it matches the
        relevant subschema of :data:`sphobjinv.schema.json_schema`, to
        facilitate implementation of
        :meth:`Inventory.json_dict()
        <sphobjinv.inventory.Inventory.json_dict>`.

        The |dict|\ s returned by :class:`~sphobjinv.data.DataObjBytes` and
        :class:`~sphobjinv.data.DataObjStr` both have |str|
        keys, but they have |bytes| and |str| values, respectively.
        The |dict| keys are identical to the |str| values of the
        :data:`~sphobjinv.data.DataFields` |Enum| members.

        Calling with both `expand` and `contract` as |True| is invalid.

        Parameters
        ----------
        expand

            |bool| *(optional)* -- Return |dict| with any
            :data:`~sphobjinv.data.SuperDataObj.uri` or
            :data:`~sphobjinv.data.SuperDataObj.dispname`
            abbreviations expanded

        contract

            |bool| *(optional)* -- Return |dict| with abbreviated
            :data:`~sphobjinv.data.SuperDataObj.uri` and
            :data:`~sphobjinv.data.SuperDataObj.dispname`

        Returns
        -------
        d

            |dict| -- Object data

        Raises
        ------
        ValueError

            If both `expand` and `contract` are |True|

        """
        if expand and contract:
            raise ValueError("'expand' and 'contract' cannot both be true.")

        d = {a: getattr(self, a) for a in (e.value for e in DataFields)}

        if expand:
            d.update(
                {
                    DataFields.URI.value: self.uri_expanded,
                    DataFields.DispName.value: self.dispname_expanded,
                }
            )

        if contract:
            d.update(
                {
                    DataFields.URI.value: self.uri_contracted,
                    DataFields.DispName.value: self.dispname_contracted,
                }
            )

        return d

    def data_line(self, *, expand=False, contract=False):
        """Compose plaintext |objects.inv| data line from instance contents.

        The format of the resulting data line is given by
        :data:`~sphobjinv.data.SuperDataObj.data_line_fmt`.
        :class:`~sphobjinv.data.DataObjBytes` and
        :class:`~sphobjinv.data.DataObjStr` instances generate data lines
        as |bytes| and |str|, respectively.

        Calling with both `expand` and `contract` as |True| is invalid.

        Parameters
        ----------
        expand

            |bool| *(optional)* -- Return data line with any
            :data:`~sphobjinv.data.SuperDataObj.uri` or
            :data:`~sphobjinv.data.SuperDataObj.dispname`
            abbreviations expanded

        contract

            |bool| *(optional)* -- Return data line with abbreviated
            :data:`~sphobjinv.data.SuperDataObj.uri` and
            :data:`~sphobjinv.data.SuperDataObj.dispname`

        Returns
        -------
        dl

            |bytes| (for :class:`~sphobjinv.data.DataObjBytes`)
            or |str| (for :class:`~sphobjinv.data.DataObjStr`)
            -- Object data line

        Raises
        ------
        ValueError

            If both `expand` and `contract` are |True|

        """
        # Rely on .json_dict to check for invalid expand == contract == True
        fmt_d = self.as_str.json_dict(expand=expand, contract=contract)

        retval = self.data_line_fmt.format(**fmt_d)

        return self._data_line_postprocess(retval)

    def evolve(self, **kwargs):
        r"""Create a new instance with changes applied.

        This helper method provides a concise means for creating new
        instances with only a subset of changed data fields.

        The names of any `kwargs` MUST be keys of the |dict|\ s
        generated by :meth:`~sphobjinv.data.SuperDataObj.json_dict`.

        Parameters
        ----------
        kwargs

            |str| or |bytes| -- Revised value(s) to use in the new
            instance for the passed keyword argument(s).

        Returns
        -------
        dobj

            :class:`~sphobjinv.data.DataObjBytes` or
            :class:`~sphobjinv.data.DataObjStr`
            -- New instance with updated data

        """
        d = self.json_dict()
        d.update(kwargs)
        return self.__class__(**d)


@attr.s(slots=True)
class DataObjStr(SuperDataObj):
    """:class:`SuperDataObj` subclass generating |str| object data.

    Two :class:`DataObjStr` instances will test equal if all of
    :attr:`~sphobjinv.data.SuperDataObj.name`,
    :attr:`~sphobjinv.data.SuperDataObj.domain`,
    :attr:`~sphobjinv.data.SuperDataObj.role`,
    :attr:`~sphobjinv.data.SuperDataObj.priority`,
    :attr:`~sphobjinv.data.SuperDataObj.uri`,
    and :attr:`~sphobjinv.data.SuperDataObj.dispname`
    are equal between them.

    .. doctest:: dataobjstr

        >>> obj = soi.DataObjStr(
        ...     name="foo",
        ...     domain="py",
        ...     role="method",
        ...     priority="1",
        ...     uri="$",
        ...     dispname="-",
        ... )
        >>> obj == obj
        True
        >>> obj == obj.evolve(name="quux")
        False

    .. versionchanged:: 2.1
        Previously, attempts to compare instances resulted in a
        :exc:`RecursionError`.

    """

    uri_abbrev = "$"
    dispname_abbrev = "-"

    name = attr.ib(converter=_utf8_decode)
    domain = attr.ib(converter=_utf8_decode)
    role = attr.ib(converter=_utf8_decode)
    priority = attr.ib(converter=_utf8_decode)
    uri = attr.ib(converter=_utf8_decode)
    dispname = attr.ib(converter=_utf8_decode)

    as_bytes = attr.ib(repr=False, eq=False)

    @as_bytes.default
    def _as_bytes_default(self):
        """Create DataObjBytes equivalent."""
        return DataObjBytes(
            name=self.name,
            domain=self.domain,
            role=self.role,
            priority=self.priority,
            uri=self.uri,
            dispname=self.dispname,
            as_str=self,
        )

    as_str = attr.ib(repr=False, eq=False)

    @as_str.default
    def _as_str_default(self):
        """Return this instance."""
        return self

    def _data_line_postprocess(self, s):
        """Perform no postprocessing."""
        return s


@attr.s(slots=True)
class DataObjBytes(SuperDataObj):
    """:class:`SuperDataObj` subclass generating |bytes| object data.

    Two :class:`DataObjBytes` instances will test equal if all of
    :attr:`~sphobjinv.data.SuperDataObj.name`,
    :attr:`~sphobjinv.data.SuperDataObj.domain`,
    :attr:`~sphobjinv.data.SuperDataObj.role`,
    :attr:`~sphobjinv.data.SuperDataObj.priority`,
    :attr:`~sphobjinv.data.SuperDataObj.uri`,
    and :attr:`~sphobjinv.data.SuperDataObj.dispname`
    are equal between them.

    .. doctest:: dataobjbytes

        >>> obj = soi.DataObjBytes(
        ...     name=b"foo",
        ...     domain=b"py",
        ...     role=b"method",
        ...     priority=b"1",
        ...     uri=b"$",
        ...     dispname=b"-",
        ... )
        >>> obj == obj
        True
        >>> obj == obj.evolve(name=b"quux")
        False

    .. versionchanged:: 2.1
        Previously, attempts to compare instances resulted in a
        :exc:`RecursionError`.

    """

    uri_abbrev = b"$"
    dispname_abbrev = b"-"

    name = attr.ib(converter=_utf8_encode)
    domain = attr.ib(converter=_utf8_encode)
    role = attr.ib(converter=_utf8_encode)
    priority = attr.ib(converter=_utf8_encode)
    uri = attr.ib(converter=_utf8_encode)
    dispname = attr.ib(converter=_utf8_encode)

    as_str = attr.ib(repr=False, eq=False)

    @as_str.default
    def _as_str_default(self):
        """Create DataObjStr equivalent."""
        return DataObjStr(
            name=self.name,
            domain=self.domain,
            role=self.role,
            priority=self.priority,
            uri=self.uri,
            dispname=self.dispname,
            as_bytes=self,
        )

    as_bytes = attr.ib(repr=False, eq=False)

    @as_bytes.default
    def _as_bytes_default(self):
        """Return this instance."""
        return self

    def _data_line_postprocess(self, s):
        """Encode to bytes before data_line return."""
        return s.encode("utf-8")



r"""*File I/O helpers for* ``sphobjinv``.

``sphobjinv`` is a toolkit for manipulation and inspection of
Sphinx |objects.inv| files.

**Author**
    Brian Skinn (brian.skinn@gmail.com)

**File Created**
    5 Nov 2017

**Copyright**
    \(c) 2016-2026 Brian Skinn and community contributors

**Source Repository**
    https://github.com/bskinn/sphobjinv

**Documentation**
    https://sphobjinv.readthedocs.io/en/stable

**License**
    Code: `MIT License`_

    Docs & Docstrings: |CC BY 4.0|_

    See |license_txt|_ for full license terms.

**Members**

"""

import json
from pathlib import Path


def readbytes(path):
    """Read file contents and return as |bytes|.

    .. versionchanged:: 2.1

        `path` can now be |Path| or |str|. Previously, it had to be |str|.

    Parameters
    ----------
    path

        |str| or |Path| -- Path to file to be opened.

    Returns
    -------
    b

        |bytes| -- Contents of the indicated file.

    """
    return Path(path).read_bytes()


def writebytes(path, contents):
    """Write indicated file contents.

    Any existing file at `path` will be overwritten.

    .. versionchanged:: 2.1

        `path` can now be |Path| or |str|. Previously, it had to be |str|.

    Parameters
    ----------
    path

        |str| or |Path| -- Path to file to be written.

    contents

        |bytes| -- Content to be written to file.

    """
    Path(path).write_bytes(contents)


def readjson(path):
    """Create |dict| from JSON file.

    No data or schema validation is performed.

    .. versionchanged:: 2.1

        `path` can now be |Path| or |str|. Previously, it had to be |str|.

    Parameters
    ----------
    path

        |str| or |Path| -- Path to JSON file to be read.

    Returns
    -------
    d

        |dict| -- Deserialized JSON.

    """
    return json.loads(Path(path).read_text())


def writejson(path, d):
    """Create JSON file from |dict|.

    No data or schema validation is performed.
    Any existing file at `path` will be overwritten.

    .. versionchanged:: 2.1

        `path` can now be |Path| or |str|. Previously, it had to be |str|.

    Parameters
    ----------
    path

        |str| or |Path| -- Path to output JSON file.

    d

        |dict| -- Data structure to serialize.

    """
    Path(path).write_text(json.dumps(d))


def urlwalk(url):
    r"""Generate a series of candidate |objects.inv| URLs.

    URLs are based on the seed `url` passed in. Ensure that the
    path separator in `url` is the standard **forward** slash
    ('|cour|\ /\ |/cour|').

    Parameters
    ----------
    url

        |str| -- Seed URL defining directory structure to walk through.

    Yields
    ------
    inv_url

        |str| -- Candidate URL for |objects.inv| location.

    """
    # Scrub any anchor, as it fouls things
    url = url.partition("#")[0]

    urlparts = url.rstrip("/").split("/")

    # This loop condition results in the yielded values stopping at
    # 'http[s]://domain.com/objects.inv', since the URL protocol
    # specifier has two forward slashes
    while len(urlparts) >= 3:
        urlparts.append("objects.inv")
        yield "/".join(urlparts)
        urlparts.pop()
        urlparts.pop()



r"""``sphobjinv`` *data class for full inventories*.

``sphobjinv`` is a toolkit for manipulation and inspection of
Sphinx |objects.inv| files.

**Author**
    Brian Skinn (brian.skinn@gmail.com)

**File Created**
    7 Dec 2017

**Copyright**
    \(c) 2016-2026 Brian Skinn and community contributors

**Source Repository**
    https://github.com/bskinn/sphobjinv

**Documentation**
    https://sphobjinv.readthedocs.io/en/stable

**License**
    Code: `MIT License`_

    Docs & Docstrings: |CC BY 4.0|_

    See |license_txt|_ for full license terms.

**Members**

"""

import re
import ssl
import urllib.request as urlrq
from zlib import error as zlib_error

import attr
import certifi
import jsonschema
from jsonschema.exceptions import ValidationError

from sphobjinv.data import DataObjStr, _utf8_encode
from sphobjinv.enum import HeaderFields, SourceTypes
from sphobjinv.fileops import readbytes
from sphobjinv.re import pb_data, pb_project, pb_version
from sphobjinv.schema import json_schema
from sphobjinv.version import __version__ as soi_version
from sphobjinv.zlib import decompress


@attr.s(slots=True, eq=True, order=False)
class Inventory:
    r"""Entire contents of an |objects.inv| inventory.

    All information is stored internally as |str|,
    even if imported from a |bytes| source.

    All arguments except `count_error` are used to specify the source
    from which the |Inventory| contents are to be populated.
    **At most ONE** of these source arguments may be other than |None|.

    The `count_error` argument is only relevant to the `dict_json` source type.

    Equality comparisons between |Inventory| instances
    will return |True| if
    :attr:`~sphobjinv.inventory.Inventory.project`,
    :attr:`~sphobjinv.inventory.Inventory.version`, and
    **all** contents of :attr:`~sphobjinv.inventory.Inventory.objects`
    are identical, even if the instances were created from different
    sources:

    .. doctest:: inventory-equality

        >>> inv1 = soi.Inventory(
        ...     url="https://sphobjinv.readthedocs.io/en/latest/objects.inv"
        ... )
        >>> inv2 = soi.Inventory(inv1.data_file())
        >>> inv1 is inv2
        False
        >>> inv1 == inv2
        True

    .. versionchanged:: 2.1
        Previously, an |Inventory| instance would compare equal only
        to itself.

    `source`

        The |Inventory| will attempt to parse the indicated
        source object as each of the below types in sequence,
        **except** for `url`.

        This argument is included mainly as a convenience
        feature for use in interactive sessions, as
        invocations of the following form implicitly
        populate `source`, as the first positional argument::

            >>> inv = Inventory(src_obj)

        In most cases, for clarity it is recommended
        that programmatic instantiation of |Inventory| objects
        utilize the below format-specific arguments.

    `plaintext`

        Object is to be parsed as the UTF-8 |bytes|
        plaintext contents of an |objects.inv| inventory.

    `zlib`

        Object is to be parsed as the UTF-8 |bytes|
        zlib-compressed contents of an
        |objects.inv| inventory.

    `fname_plain`

        Object is the |str| or |Path| path to a file containing
        the plaintext contents of an |objects.inv| inventory.

        .. versionchanged:: 2.1

            Previously, this argument could only be a |str|.

    `fname_zlib`

        Object is the |str| or |Path| path to a file containing
        the zlib-compressed contents of an
        |objects.inv| inventory.

        .. versionchanged:: 2.1

            Previously, this argument could only be a |str|.

    `dict_json`

        Object is a |dict| containing the contents
        of an |objects.inv| inventory, conforming to
        the JSON schema of
        :data:`schema.json_schema <sphobjinv.schema.json_schema>`.

        If `count_error` is passed as |True|,
        then a :exc:`ValueError` is raised if
        the number of objects found in the |dict|
        does not match the value associated with
        its `count` key.
        If `count_error` is passed as |False|,
        an object count mismatch is ignored.

    `url`

        Object is a |str| URL to a zlib-compressed
        |objects.inv| file.
        Any URL type supported by
        :mod:`urllib.request` SHOULD work; only
        |cour|\ http:\ |/cour| and
        |cour|\ file:\ |/cour|
        have been directly tested, however.

        No authentication is supported at this time.

    **Members**

    """

    # General source for try-most-types import
    # Needs to be first so it absorbs a positional arg
    _source = attr.ib(repr=False, default=None, eq=False)

    # Stringlike types (both accept str & bytes)
    _plaintext = attr.ib(repr=False, default=None, eq=False)
    _zlib = attr.ib(repr=False, default=None, eq=False)

    # Filename types (must be str or Path)
    _fname_plain = attr.ib(repr=False, default=None, eq=False)
    _fname_zlib = attr.ib(repr=False, default=None, eq=False)

    # dict types
    _dict_json = attr.ib(repr=False, default=None, eq=False)

    # URL for remote retrieval of objects.inv/.txt
    _url = attr.ib(repr=False, default=None, eq=False)

    # Flag for whether to raise error on object count mismatch
    _count_error = attr.ib(
        repr=False, default=True, validator=attr.validators.instance_of(bool), eq=False
    )

    # Actual regular attributes
    #: |str| project display name for the inventory
    #: (see :ref:`here <syntax-mouseover-example>`).
    project = attr.ib(init=False, default=None)

    #: |str| project display version for the inventory
    #: (see :ref:`here <syntax-mouseover-example>`).
    version = attr.ib(init=False, default=None)

    #: |list| of |DataObjStr| representing the
    #: data objects of the inventory.
    #: Can be edited directly to change the inventory contents.
    #: Undefined/random behavior/errors will result if the type
    #: of the elements is anything other than |DataObjStr|.
    objects = attr.ib(init=False, default=attr.Factory(list), repr=False)

    #: :class:`~sphobjinv.enum.SourceTypes` |Enum| value indicating the type of
    #: source from which the instance was generated.
    source_type = attr.ib(init=False, default=None, eq=False)

    # Helper strings for inventory datafile output
    #: Preamble line for v2 |objects.inv| header
    header_preamble = "# Sphinx inventory version 2"

    #: Project line |str.format| template for |objects.inv| header
    header_project = "# Project: {project}"

    #: Version line |str.format| template for |objects.inv| header
    header_version = "# Version: {version}"

    #: zlib compression line for v2 |objects.inv| header
    header_zlib = "# The remainder of this file is compressed using zlib."

    # Private class member for SSL context, since context creation is slow(?)
    _sslcontext = ssl.create_default_context(cafile=certifi.where())

    @property
    def count(self):
        """Count of objects currently in inventory."""
        return len(self.objects)

    def json_dict(self, expand=False, contract=False):
        """Generate a flat |dict| representation of the inventory.

        The returned |dict| matches the
        schema of :data:`sphobjinv.schema.json_schema`.

        Calling with both `expand` and `contract` as |True| is invalid.

        Parameters
        ----------
        expand

            |bool| *(optional)* -- Return |dict| with any
            :data:`~sphobjinv.data.SuperDataObj.uri` or
            :data:`~sphobjinv.data.SuperDataObj.dispname`
            abbreviations expanded

        contract

            |bool| *(optional)* -- Return |dict| with abbreviated
            :data:`~sphobjinv.data.SuperDataObj.uri` and
            :data:`~sphobjinv.data.SuperDataObj.dispname` values

        Returns
        -------
        d

            |dict| -- Inventory data; keys and values are all |str|

        Raises
        ------
        ValueError

            If both `expand` and `contract` are |True|

        """
        d = {
            HeaderFields.Project.value: self.project,
            HeaderFields.Version.value: self.version,
            HeaderFields.Count.value: self.count,
        }

        for i, o in enumerate(self.objects):
            d.update({str(i): o.json_dict(expand=expand, contract=contract)})

        return d

    @property
    def objects_rst(self):
        r"""|list| of objects formatted in a |str| reST-like representation.

        The format of each |str| in the |list| is given by
        :class:`data.SuperDataObj.rst_fmt
        <sphobjinv.data.SuperDataObj.rst_fmt>`.

        Returns
        -------
        obj_l

            |list| of |str| -- Inventory object data in reST-like format

        """
        return [_.as_rst for _ in self.objects]

    def __str__(self):  # pragma: no cover
        """Return concise, readable description of contents."""
        ret_str = "<{0} ({1}): {2} {3}, {4} objects>"

        proj = self.project if self.project else "<no project>"
        ver = "v" + self.version if self.version else "<no version>"

        return ret_str.format(
            type(self).__name__, self.source_type.value, proj, ver, self.count
        )

    def __attrs_post_init__(self):
        """Construct the inventory from the indicated source."""
        # List of sources
        src_list = (
            self._source,
            self._plaintext,
            self._zlib,
            self._fname_plain,
            self._fname_zlib,
            self._dict_json,
            self._url,
        )
        src_count = sum(1 for _ in src_list if _ is not None)

        # Complain if multiple sources provided
        if src_count > 1:
            raise RuntimeError("At most one data source can be specified.")

        # Leave uninitialized ("manual" init) if no source provided
        if src_count == 0:
            self.source_type = SourceTypes.Manual
            return

        # If general ._source was provided, run the generalized import
        if self._source is not None:
            self._general_import()
            return

        # For all of these below, '()' is passed as 'exc' argument since
        # desire _try_import not to handle any exception types

        # Plaintext str or bytes
        # Special case, since preconverting input.
        if self._plaintext is not None:
            self._try_import(
                self._import_plaintext_bytes, _utf8_encode(self._plaintext), ()
            )
            self.source_type = SourceTypes.BytesPlaintext
            return

        # Remainder are iterable
        for src, fxn, st in zip(
            (
                self._zlib,
                self._fname_plain,
                self._fname_zlib,
                self._dict_json,
                self._url,
            ),
            (
                self._import_zlib_bytes,
                self._import_plaintext_fname,
                self._import_zlib_fname,
                self._import_json_dict,
                self._import_url,
            ),
            (
                SourceTypes.BytesZlib,
                SourceTypes.FnamePlaintext,
                SourceTypes.FnameZlib,
                SourceTypes.DictJSON,
                SourceTypes.URL,
            ),
        ):
            if src is not None:
                self._try_import(fxn, src, ())
                self.source_type = st
                return

    def data_file(self, *, expand=False, contract=False):
        """Generate a plaintext |objects.inv| as UTF-8 |bytes|.

        |bytes| is used here as the output type
        since the most common use cases are anticipated to be
        either (1) dumping to file via :func:`sphobjinv.fileops.writebytes`
        or (2) compressing via :func:`sphobjinv.zlib.compress`,
        both of which take |bytes| input.

        Calling with both `expand` and `contract` as |True| is invalid.

        Parameters
        ----------
        expand

            |bool| *(optional)* -- Generate |bytes| with any
            :data:`~sphobjinv.data.SuperDataObj.uri` or
            :data:`~sphobjinv.data.SuperDataObj.dispname`
            abbreviations expanded

        contract

            |bool| *(optional)* -- Generate |bytes| with abbreviated
            :data:`~sphobjinv.data.SuperDataObj.uri` and
            :data:`~sphobjinv.data.SuperDataObj.dispname` values

        Returns
        -------
        b

            |bytes| -- Inventory in plaintext |objects.inv| format

        Raises
        ------
        ValueError

            If both `expand` and `contract` are |True|

        """
        # Rely on SuperDataObj to proof expand/contract args
        # Extra empty string at the end puts a newline at the end
        # of the generated string, consistent with files
        # generated by Sphinx.
        return "\n".join(
            (
                self.header_preamble,
                self.header_project.format(project=self.project),
                self.header_version.format(version=self.version),
                self.header_zlib,
                *(
                    obj.data_line(expand=expand, contract=contract)
                    for obj in self.objects
                ),
                "",
            )
        ).encode("utf-8")

    def suggest(self, name, *, thresh=50, with_index=False, with_score=False):
        r"""Suggest objects in the inventory to match a name.

        :meth:`~Inventory.suggest` makes use of
        the edit-distance scoring library |fuzzywuzzy|_
        to identify potential matches to the given `name`
        within the inventory.
        The search is performed over the |list| of |str|
        generated by :meth:`~objects_rst`.

        `thresh` defines the minimum |fuzzywuzzy|_ match quality
        (an integer ranging from 0 to 100)
        required for a given object to be included
        in the results list.
        Can be any float value,
        but best results are generally obtained
        with values between 50 and 80,
        depending on the number of objects in the inventory,
        the confidence of the user in the match
        between `name` and the object(s) of interest,
        and the desired fidelity of the search results to `name`.

        This functionality is provided by the
        :doc:`'suggest' subparser </cli/suggest>`
        of the command-line interface.

        Parameters
        ----------
        name

            |str| -- Object name for |fuzzywuzzy|_ pattern matching

        thresh

            |float| -- |fuzzywuzzy|_ match quality threshold

        with_index

            |bool| -- Include with each matched name
            its index within :attr:`Inventory.objects`

        with_score

            |bool| -- Include with each matched name
            its |fuzzywuzzy|_ match quality score

        Returns
        -------
        res_l

            |list|

            If both `with_index` and `with_score`
            are |False|, members are the |str|
            :meth:`SuperDataObj.as_rst()
            <sphobjinv.data.SuperDataObj.as_rst>`
            representations of matching objects.

            If either is |True|, members are |tuple|\ s
            of the indicated match results:

            `with_index == True`: |cour|\ (as_rst, index)\ |/cour|

            `with_score == True`: |cour|\ (as_rst, score)\ |/cour|

            `with_index == with_score == True`:
            |cour|\ (as_rst, score, index)\ |/cour|

        """
        from sphobjinv._vendored.fuzzywuzzy import process as fwp

        # Must propagate list index to include in output
        # Search vals are rst prepended with list index
        srch_list = [f"{i} {o}" for i, o in enumerate(self.objects_rst)]

        # Composite each string result extracted by fuzzywuzzy
        # and its match score into a single string. The match
        # and score are returned together in a tuple.
        results = [
            f"{match} {score}"
            for match, score in fwp.extract(name, srch_list, limit=None)
            if score >= thresh
        ]

        # Define regex for splitting the three components, and
        # use it to convert composite result string to tuple:
        # result --> (rst, score, index)
        p_idx = re.compile(r"^(\d+)\s+(.+?)\s+(\d+)$")
        results = [
            (m.group(2), int(m.group(3)), int(m.group(1)))
            for m in map(p_idx.match, results)
        ]

        # Return based on flags
        if with_score:
            if with_index:
                return results
            else:
                return [tup[:2] for tup in results]
        else:
            if with_index:
                return [tup[::2] for tup in results]
            else:
                return [tup[0] for tup in results]

    def _general_import(self):
        """Attempt sequence of all imports."""
        # Lookups for method names and expected import-failure errors
        importers = {
            SourceTypes.BytesPlaintext: self._import_plaintext_bytes,
            SourceTypes.BytesZlib: self._import_zlib_bytes,
            SourceTypes.FnamePlaintext: self._import_plaintext_fname,
            SourceTypes.FnameZlib: self._import_zlib_fname,
            SourceTypes.DictJSON: self._import_json_dict,
        }
        import_errors = {
            SourceTypes.BytesPlaintext: TypeError,
            SourceTypes.BytesZlib: (zlib_error, TypeError),
            SourceTypes.FnamePlaintext: (OSError, TypeError, UnicodeDecodeError),
            SourceTypes.FnameZlib: (OSError, TypeError, zlib_error),
            SourceTypes.DictJSON: (ValidationError),
        }

        # Attempt series of import approaches
        # Enum keys are ordered, so iteration is too.
        for st in SourceTypes:
            if st not in importers:
                # No action for source types w/o a handler function defined.
                continue

            if self._try_import(importers[st], self._source, import_errors[st]):
                self.source_type = st
                return

        # Nothing worked, complain.
        raise TypeError("Invalid Inventory source type")

    def _try_import(self, import_fxn, src, exc):
        """Attempt the indicated import method on the indicated source.

        Returns True on success.

        """
        try:
            p, v, o = import_fxn(src)
        except exc:
            return False

        self.project = p
        self.version = v
        self.objects = o

        return True

    def _import_plaintext_bytes(self, b_str):
        """Import an inventory from plaintext UTF-8 bytes."""
        b_res = pb_project.search(b_str).group(HeaderFields.Project.value)
        project = b_res.decode("utf-8")

        b_res = pb_version.search(b_str).group(HeaderFields.Version.value)
        version = b_res.decode("utf-8")

        def gen_dataobjs():
            """Generate a data object for each line in the inventory."""
            for mch in pb_data.finditer(b_str):
                yield DataObjStr(**mch.groupdict())

        objects = []
        objects.extend(gen_dataobjs())

        if len(objects) == 0:
            raise TypeError("No objects found in plaintext")

        return project, version, objects

    def _import_zlib_bytes(self, b_str):
        """Import a zlib-compressed inventory."""
        b_plain = decompress(b_str)
        p, v, o = self._import_plaintext_bytes(b_plain)

        return p, v, o

    def _import_plaintext_fname(self, fn):
        """Import a plaintext inventory file."""
        b_plain = readbytes(fn)

        return self._import_plaintext_bytes(b_plain)

    def _import_zlib_fname(self, fn):
        """Import a zlib-compressed inventory file."""
        b_zlib = readbytes(fn)

        return self._import_zlib_bytes(b_zlib)

    def _import_url(self, url):
        """Import a file from a remote URL."""
        # Caller's responsibility to ensure URL points
        # someplace safe/sane!
        req = urlrq.Request(url, headers={"User-Agent": "sphobjinv URL/" + soi_version})
        resp = urlrq.urlopen(req, context=self._sslcontext)  # noqa: S310
        b_str = resp.read()

        # Plaintext URL D/L is unreliable; zlib only
        return self._import_zlib_bytes(b_str)

    def _import_json_dict(self, d):
        """Import flat-dict composited data."""
        # Validate the dict against the schema. Schema
        # WILL allow an inventory with no objects here
        val = jsonschema.Draft4Validator(json_schema)
        val.validate(d)

        # Pull header items first
        project = d[HeaderFields.Project.value]
        version = d[HeaderFields.Version.value]
        count = d[HeaderFields.Count.value]

        # No objects is not allowed
        if count < 1:
            raise ValueError("Import of zero-length inventory")

        # Going to destructively process d, so shallow-copy it first
        d = d.copy()

        # Expecting the dict to be indexed by string integers
        objects = []
        for i in range(count):
            try:
                objects.append(DataObjStr(**d.pop(str(i))))
            except KeyError as e:
                if self._count_error:
                    err_str = (
                        f"Too few objects found in dict (halt at {i}, expect {count})"
                    )
                    raise ValueError(err_str) from e

        # Complain if remaining objects are anything other than the
        # valid inventory-level header keys
        hf_values = {e.value for e in HeaderFields}
        check_value = self._count_error and set(d.keys()).difference(hf_values)
        if check_value:
            # A truthy value here will be the contents
            # of the above set difference
            err_str = f"Too many objects in dict ({check_value})"
            raise ValueError(err_str)

        # Should be good to return
        return project, version, objects



r"""*Custom errors for* ``sphobjinv``.

``sphobjinv`` is a toolkit for manipulation and inspection of
Sphinx |objects.inv| files.

**Author**
    Brian Skinn (brian.skinn@gmail.com)

**File Created**
    5 Nov 2017

**Copyright**
    \(c) 2016-2026 Brian Skinn and community contributors

**Source Repository**
    https://github.com/bskinn/sphobjinv

**Documentation**
    https://sphobjinv.readthedocs.io/en/stable

**License**
    Code: `MIT License`_

    Docs & Docstrings: |CC BY 4.0|_

    See |license_txt|_ for full license terms.

**Members**

"""


class SphobjinvError(Exception):
    """Custom ``sphobjinv`` error superclass."""


class VersionError(SphobjinvError):
    """Raised when attempting an operation on an unsupported version.

    The current version of ``sphobjinv`` only supports 'version 2'
    |objects.inv| files (see :doc:`here </syntax>`).

    """

    # TODO: Add SOI prefix to this class name as part of the exceptions refactor



# -*- coding: utf8 -*-

from sphobjinv._vendored.fuzzywuzzy.fuzz import *
from sphobjinv._vendored.fuzzywuzzy import process
from sphobjinv._vendored.fuzzywuzzy import utils

import itertools
import unittest

class UtilsTest(unittest.TestCase):
    def setUp(self):
        self.s1 = "new york mets"
        self.s1a = "new york mets"
        self.s2 = "new YORK mets"
        self.s3 = "the wonderful new york mets"
        self.s4 = "new york mets vs atlanta braves"
        self.s5 = "atlanta braves vs new york mets"
        self.s6 = "new york mets - atlanta braves"
        self.mixed_strings = [
            "Lorem Ipsum is simply dummy text of the printing and typesetting industry.",
            "C'est la vie",
            "Ça va?",
            "Cães danados",
            u"\xacCamarões assados",
            u"a\xac\u1234\u20ac\U00008000",
            u"\u00C1"
        ]


    def tearDown(self):
        pass

    def test_asciidammit(self):
        for s in self.mixed_strings:
            utils.asciidammit(s)

    def test_asciionly(self):
        for s in self.mixed_strings:
            # ascii only only runs on strings
            s = utils.asciidammit(s)
            utils.asciionly(s)

    def test_fullProcess(self):
        for s in self.mixed_strings:
            utils.full_process(s)

class RatioTest(unittest.TestCase):

    def setUp(self):
        self.s1 = "new york mets"
        self.s1a = "new york mets"
        self.s2 = "new YORK mets"
        self.s3 = "the wonderful new york mets"
        self.s4 = "new york mets vs atlanta braves"
        self.s5 = "atlanta braves vs new york mets"
        self.s6 = "new york mets - atlanta braves"

        self.cirque_strings = [
            "cirque du soleil - zarkana - las vegas",
            "cirque du soleil ",
            "cirque du soleil las vegas",
            "zarkana las vegas",
            "las vegas cirque du soleil at the bellagio",
            "zarakana - cirque du soleil - bellagio"
        ]

        self.baseball_strings = [
            "new york mets vs chicago cubs",
            "chicago cubs vs chicago white sox",
            "philladelphia phillies vs atlanta braves",
            "braves vs mets",
        ]

    def tearDown(self):
        pass

    def testEqual(self):
        self.assertEqual(ratio(self.s1, self.s1a),100)

    def testCaseInsensitive(self):
        self.assertNotEqual(ratio(self.s1, self.s2),100)
        self.assertEqual(ratio(utils.full_process(self.s1), utils.full_process(self.s2)),100)

    def testPartialRatio(self):
        self.assertEqual(partial_ratio(self.s1, self.s3),100)

    def testTokenSortRatio(self):
        self.assertEqual(token_sort_ratio(self.s1, self.s1a),100)

    def testPartialTokenSortRatio(self):
        self.assertEqual(partial_token_sort_ratio(self.s1, self.s1a),100)
        self.assertEqual(partial_token_sort_ratio(self.s4, self.s5),100)

    def testTokenSetRatio(self):
        self.assertEqual(token_set_ratio(self.s4, self.s5),100)

    def testPartialTokenSetRatio(self):
        self.assertEqual(partial_token_set_ratio(self.s4, self.s5),100)

    def testQuickRatioEqual(self):
        self.assertEqual(QRatio(self.s1, self.s1a), 100)

    def testQuickRatioCaseInsensitive(self):
        self.assertEqual(QRatio(self.s1, self.s2), 100)

    def testQuickRatioNotEqual(self):
        self.assertNotEqual(QRatio(self.s1, self.s3), 100)

    def testWRatioEqual(self):
        self.assertEqual(WRatio(self.s1, self.s1a), 100)

    def testWRatioCaseInsensitive(self):
        self.assertEqual(WRatio(self.s1, self.s2), 100)

    def testWRatioPartialMatch(self):
        # a partial match is scaled by .9
        self.assertEqual(WRatio(self.s1, self.s3), 90)

    def testWRatioMisorderedMatch(self):
        # misordered full matches are scaled by .95
        self.assertEqual(WRatio(self.s4, self.s5), 95)

    def testWRatioUnicode(self):
        self.assertEqual(WRatio(unicode(self.s1), unicode(self.s1a)), 100)

    def testQRatioUnicode(self):
        self.assertEqual(WRatio(unicode(self.s1), unicode(self.s1a)), 100)

    def testIssueSeven(self):
        s1 = "HSINCHUANG"
        s2 = "SINJHUAN"
        s3 = "LSINJHUANG DISTRIC"
        s4 = "SINJHUANG DISTRICT"

        self.assertGreater(partial_ratio(s1, s2), 75)
        self.assertGreater(partial_ratio(s1, s3), 75)
        self.assertGreater(partial_ratio(s1, s4), 75)

    def testWRatioUnicodeString(self):
        s1 = u"\u00C1"
        s2 = "ABCD"
        score = WRatio(s1, s2)
        self.assertEqual(0, score)

    def testQRatioUnicodeString(self):
        s1 = u"\u00C1"
        s2 = "ABCD"
        score = QRatio(s1, s2)
        self.assertEqual(0, score)

    # test processing methods
    def testGetBestChoice1(self):
        query = "new york mets at atlanta braves"
        best = process.extractOne(query, self.baseball_strings)
        self.assertEqual(best[0], "braves vs mets")

    def testGetBestChoice2(self):
        query = "philadelphia phillies at atlanta braves"
        best = process.extractOne(query, self.baseball_strings)
        self.assertEqual(best[0], self.baseball_strings[2])

    def testGetBestChoice3(self):
        query = "atlanta braves at philadelphia phillies"
        best = process.extractOne(query, self.baseball_strings)
        self.assertEqual(best[0], self.baseball_strings[2])

    def testGetBestChoice4(self):
        query = "chicago cubs vs new york mets"
        best = process.extractOne(query, self.baseball_strings)
        self.assertEqual(best[0], self.baseball_strings[0])

class ProcessTest(unittest.TestCase):

    def setUp(self):
        self.s1 = "new york mets"
        self.s1a = "new york mets"
        self.s2 = "new YORK mets"
        self.s3 = "the wonderful new york mets"
        self.s4 = "new york mets vs atlanta braves"
        self.s5 = "atlanta braves vs new york mets"
        self.s6 = "new york mets - atlanta braves"

        self.cirque_strings = [
            "cirque du soleil - zarkana - las vegas",
            "cirque du soleil ",
            "cirque du soleil las vegas",
            "zarkana las vegas",
            "las vegas cirque du soleil at the bellagio",
            "zarakana - cirque du soleil - bellagio"
        ]

        self.baseball_strings = [
            "new york mets vs chicago cubs",
            "chicago cubs vs chicago white sox",
            "philladelphia phillies vs atlanta braves",
            "braves vs mets",
        ]

    def testWithProcessor(self):
        events = [
            ["chicago cubs vs new york mets", "CitiField", "2011-05-11", "8pm"],
            ["new york yankees vs boston red sox", "Fenway Park", "2011-05-11", "8pm"],
            ["atlanta braves vs pittsburgh pirates", "PNC Park", "2011-05-11", "8pm"],
        ]
        query = "new york mets vs chicago cubs"
        processor = lambda event: event[0]

        best = process.extractOne(query, events, processor=processor)
        self.assertEqual(best[0], events[0])

    def testWithScorer(self):
        choices = [
            "new york mets vs chicago cubs",
            "chicago cubs at new york mets",
            "atlanta braves vs pittsbugh pirates",
            "new york yankees vs boston red sox"
        ]

        # in this hypothetical example we care about ordering, so we use quick ratio
        query = "new york mets at chicago cubs"
        scorer = QRatio

        # first, as an example, the normal way would select the "more 'complete' match of choices[1]"

        best = process.extractOne(query, choices)
        self.assertEqual(best[0], choices[1])

        # now, use the custom scorer

        best = process.extractOne(query, choices, scorer=scorer)
        self.assertEqual(best[0], choices[0])

    def testWithCutoff(self):
        choices = [
            "new york mets vs chicago cubs",
            "chicago cubs at new york mets",
            "atlanta braves vs pittsbugh pirates",
            "new york yankees vs boston red sox"
        ]

        query = "los angeles dodgers vs san francisco giants"

        # in this situation, this is an event that does not exist in the list
        # we don't want to randomly match to something, so we use a reasonable cutoff

        best = process.extractOne(query, choices, score_cutoff=50)
        self.assertTrue(best is None)
        #self.assertIsNone(best) # unittest.TestCase did not have assertIsNone until Python 2.7

        # however if we had no cutoff, something would get returned

        #best = process.extractOne(query, choices)
        #self.assertIsNotNone(best)

    def testEmptyStrings(self):
        choices = [
            "",
            "new york mets vs chicago cubs",
            "new york yankees vs boston red sox",
            "",
            ""
        ]

        query = "new york mets at chicago cubs"

        best = process.extractOne(query, choices)
        self.assertEqual(best[0], choices[1])

    def testNullStrings(self):
        choices = [
            None,
            "new york mets vs chicago cubs",
            "new york yankees vs boston red sox",
            None,
            None
        ]

        query = "new york mets at chicago cubs"

        best = process.extractOne(query, choices)
        self.assertEqual(best[0], choices[1])


if __name__ == '__main__':
    unittest.main()         # run all tests


r"""``sphobjinv._vendored.fuzzywuzzy`` *package definition module*.

``sphobjinv`` is a toolkit for manipulation and inspection of
Sphinx |objects.inv| files.

This subpackage vendors an archival version of fuzzywuzzy from
before it incorporated python-Levenshtein and thus acquired
the GPL licence by flow-through, overriding the prior
MIT license of the package.

All code in this subpackage vendored from:
https://github.com/seatgeek/fuzzywuzzy/tree/4bf28161f7005f3aa9d4d931455ac55126918df7

Imports and code details were changed as required to make the code compatible
with current (2021) versions of Python 3.

Link to MIT-licensed fuzzywuzzy version obtained from the bottom of the README at:
https://github.com/maxbachmann/RapidFuzz/tree/460d291a38922c7fead920de8147798b817653cb


**Author**
    Brian Skinn (brian.skinn@gmail.com)

**File Created**
    11 Dec 2021

**Copyright**
    \(c) 2016-2026 Brian Skinn and community contributors

**Source Repository**
    https://github.com/bskinn/sphobjinv

**Documentation**
    https://sphobjinv.readthedocs.io/en/stable

**License**
    Code: `MIT License`_

    Docs & Docstrings: |CC BY 4.0|_

    See |license_txt|_ for full license terms.

**Members**

"""



r"""*zlib (de)compression helpers for* ``sphobjinv``.

``sphobjinv`` is a toolkit for manipulation and inspection of
Sphinx |objects.inv| files.

**Author**
    Brian Skinn (brian.skinn@gmail.com)

**File Created**
    5 Nov 2017

**Copyright**
    \(c) 2016-2026 Brian Skinn and community contributors

**Source Repository**
    https://github.com/bskinn/sphobjinv

**Documentation**
    https://sphobjinv.readthedocs.io/en/stable

**License**
    Code: `MIT License`_

    Docs & Docstrings: |CC BY 4.0|_

    See |license_txt|_ for full license terms.

**Members**

"""

import io
import os
import zlib

BUFSIZE = 16 * 1024  # 16k chunks


def decompress(bstr):
    """Decompress a version 2 |isphx| |objects.inv| bytestring.

    The `#`-prefixed comment lines are left unchanged, whereas the
    :mod:`zlib`-compressed data lines are decompressed to plaintext.

    Parameters
    ----------
    bstr

        |bytes| -- Binary string containing a compressed |objects.inv|
        file.

    Returns
    -------
    out_b

        |bytes| -- Decompressed binary string containing the plaintext
        |objects.inv| content.

    """
    from sphobjinv.error import VersionError

    def decompress_chunks(bstrm):
        """Handle chunk-wise zlib decompression.

        Internal function pulled from intersphinx.py@v1.4.1:
        https://github.com/sphinx-doc/sphinx/blob/1.4.1/sphinx/
        ext/intersphinx.py#L79-L124.

        BUFSIZE taken as the default value from intersphinx signature
        Modified slightly to take the stream as a parameter,
        rather than assuming one from the parent namespace.

        """
        decompressor = zlib.decompressobj()
        for chunk in iter(lambda: bstrm.read(BUFSIZE), b""):
            yield decompressor.decompress(chunk)
        yield decompressor.flush()

    # Make stream and output string
    strm = io.BytesIO(bstr)

    # Check to be sure it's v2
    out_b = strm.readline()
    if not out_b.endswith(b"2\n"):  # pragma: no cover
        raise VersionError("Only v2 objects.inv files currently supported")

    # Pull name, version, and description lines
    for _ in range(3):
        out_b += strm.readline()

    # Decompress chunks and append
    for chunk in decompress_chunks(strm):
        out_b += chunk

    # Replace newlines with the OS-local newlines, and return
    return out_b.replace(b"\n", os.linesep.encode("utf-8"))


def compress(bstr):
    """Compress a version 2 |isphx| |objects.inv| bytestring.

    The `#`-prefixed comment lines are left unchanged, whereas the
    plaintext data lines are compressed with :mod:`zlib`.

    Parameters
    ----------
    bstr

        |bytes| -- Binary string containing the decompressed contents of an
        |objects.inv| file.

    Returns
    -------
    out_b

        |bytes| -- Binary string containing the compressed |objects.inv|
        content.

    """
    from sphobjinv.re import pb_comments, pb_data

    # Preconvert any DOS newlines to Unix
    s = bstr.replace(b"\r\n", b"\n")

    # Pull all of the lines
    m_comments = pb_comments.findall(s)
    m_data = pb_data.finditer(s)

    # Assemble the binary header comments and data
    # Comments and data blocks must end in newlines
    hb = b"\n".join(m_comments) + b"\n"
    db = b"\n".join(_.group(0) for _ in m_data) + b"\n"

    # Compress the data block
    # Compression level nine is to match that specified in
    #  sphinx html builder:
    # https://github.com/sphinx-doc/sphinx/blob/1.4.1/sphinx/
    #    builders/html.py#L843
    dbc = zlib.compress(db, 9)

    # Return the composited bytestring
    return hb + dbc



r"""``sphobjinv`` *package definition module*.

``sphobjinv`` is a toolkit for manipulation and inspection of
Sphinx |objects.inv| files.

**Author**
    Brian Skinn (brian.skinn@gmail.com)

**File Created**
    17 May 2016

**Copyright**
    \(c) 2016-2026 Brian Skinn and community contributors

**Source Repository**
    https://github.com/bskinn/sphobjinv

**Documentation**
    https://sphobjinv.readthedocs.io/en/stable

**License**
    Code: `MIT License`_

    Docs & Docstrings: |CC BY 4.0|_

    See |license_txt|_ for full license terms.

**Members**

"""

from sphobjinv.data import DataFields, DataObjBytes, DataObjStr
from sphobjinv.enum import HeaderFields, SourceTypes
from sphobjinv.error import SphobjinvError, VersionError
from sphobjinv.fileops import readbytes, readjson, urlwalk, writebytes, writejson
from sphobjinv.inventory import Inventory
from sphobjinv.re import p_data, pb_comments, pb_data, pb_project, pb_version
from sphobjinv.schema import json_schema
from sphobjinv.version import __version__
from sphobjinv.zlib import compress, decompress



#!/usr/bin/env python
# encoding: utf-8
"""
process.py

Copyright (c) 2011 Adam Cohen

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
from sphobjinv._vendored.fuzzywuzzy.fuzz import *

import sys, os
from sphobjinv._vendored.fuzzywuzzy import utils

#######################################
# Find Best Matchs In List Of Choices #
#######################################

def extract(query, choices, processor=None, scorer=None, limit=5):

    # choices       = a list of objects we are attempting to extract values from
    # query         = an object representing the thing we want to find
    # scorer        f(OBJ, QUERY) --> INT. We will return the objects with the highest score
        # by default, we use score.WRatio() and both OBJ and QUERY should be strings
    # processor     f(OBJ_A) --> OBJ_B, where the output is an input to scorer
        # for example, "processor = lambda x: x[0]" would return the first element in a collection x (of, say, strings)
        # this would then be used in the scoring collection

    if choices is None or len(choices) == 0:
        return []

    # default, turn whatever the choice is into a string
    if processor is None:
        processor = lambda x: utils.asciidammit(x)

    # default: wratio
    if scorer is None:
        scorer = WRatio

    sl = list()

    for choice in choices:
        processed = processor(choice)
        score = scorer(query, processed)
        tuple = (choice, score)
        sl.append(tuple)

    sl.sort(key=lambda i: -1*i[1])
    return sl[:limit]

##########################
# Find Single Best Match #
##########################

def extractOne(query, choices, processor=None, scorer=None, score_cutoff=0):

    # convenience method which returns the single best choice
    # optional parameter: score_cutoff.
        # If the best choice has a score of less than score_cutoff
        # we will return none (intuition: not a good enough match)

    best_list = extract(query, choices, processor, scorer, limit=1)
    if len(best_list) > 0:
        best = best_list[0]
        if best[1] > score_cutoff:
            return best
        else:
            return None
    else:
        return None



r"""``sphobjinv`` *module for CLI convert functionality*.

``sphobjinv`` is a toolkit for manipulation and inspection of
Sphinx |objects.inv| files.

**Author**
    Brian Skinn (brian.skinn@gmail.com)

**File Created**
    20 Oct 2022

**Copyright**
    \(c) 2016-2026 Brian Skinn and community contributors

**Source Repository**
    https://github.com/bskinn/sphobjinv

**Documentation**
    https://sphobjinv.readthedocs.io/en/stable

**License**
    Code: `MIT License`_

    Docs & Docstrings: |CC BY 4.0|_

    See |license_txt|_ for full license terms.

**Members**

"""

from sphobjinv.cli.parser import PrsConst
from sphobjinv.cli.write import write_file, write_stdout


def do_convert(inv, in_path, params):
    r"""Carry out the conversion operation, including writing output.

    If OVERWRITE is passed and the output file
    (the default location, or as passed to OUTFILE)
    exists, it will be overwritten without a prompt. Otherwise,
    the user will be queried if it is desired to overwrite
    the existing file.

    If QUIET is passed, nothing will be
    printed to |cour|\ stdout\ |/cour|
    (potentially useful for scripting),
    and any existing output file will be overwritten
    without prompting.

    Parameters
    ----------
    inv

        |Inventory| -- Inventory object to be output in the format
        indicated by MODE.

    in_path

        |str| -- For a local input file, its absolute path.
        For a URL, the (possibly truncated) URL text.

    params

        |dict| -- Parameters/values mapping from the active subparser

    """
    if params[PrsConst.OUTFILE] == "-" or (
        params[PrsConst.INFILE] == "-" and params[PrsConst.OUTFILE] is None
    ):
        write_stdout(inv, params)
    else:
        write_file(inv, in_path, params)



r"""``sphobjinv`` *package execution module*.

``sphobjinv`` is a toolkit for manipulation and inspection of
Sphinx |objects.inv| files.

**Author**
    Brian Skinn (brian.skinn@gmail.com)

**File Created**
    15 May 2020

**Copyright**
    \(c) 2016-2026 Brian Skinn and community contributors

**Source Repository**
    https://github.com/bskinn/sphobjinv

**Documentation**
    https://sphobjinv.readthedocs.io/en/stable

**License**
    Code: `MIT License`_

    Docs & Docstrings: |CC BY 4.0|_

    See |license_txt|_ for full license terms.

**Members**

"""

import sys

from sphobjinv.cli.core import main

if __name__ == "__main__":
    # Spoof so 'help' usage display shows "sphobjinv" and
    # not "__main__.py"
    sys.argv[0] = "sphobjinv"

    sys.exit(main())



import string

bad_chars=b""  # B Skinn 2021-12-11
for i in range(128,256):
    bad_chars+=chr(i).encode()  # B Skinn 2021-12-11
table_from=string.punctuation+string.ascii_uppercase
table_to=' '*len(string.punctuation)+string.ascii_lowercase
trans_table=bytes.maketrans(table_from.encode(), table_to.encode())  # B Skinn 2021-12-11


def asciionly(s):
    return s.encode().translate(None, bad_chars).decode(errors='replace')  # B Skinn 2021-12-11

# remove non-ASCII characters from strings
def asciidammit(s):
    if type(s) is str:
        return asciionly(s)
    elif type(s) is unicode:
        return asciionly(s.encode('ascii', 'ignore'))
    else:
        return asciidammit(unicode(s))

def validate_string(s):
    try:
        if len(s)>0:
            return True
        else:
            return False
    except:
        return False

def full_process(s):
    s = asciidammit(s)
    # B Skinn 2021-12-11
    return s.encode().translate(trans_table, bad_chars).decode(errors='replace').strip()



r"""*Helper regexes for* ``sphobjinv``.

``sphobjinv`` is a toolkit for manipulation and inspection of
Sphinx |objects.inv| files.

**Author**
    Brian Skinn (brian.skinn@gmail.com)

**File Created**
    5 Nov 2017

**Copyright**
    \(c) 2016-2026 Brian Skinn and community contributors

**Source Repository**
    https://github.com/bskinn/sphobjinv

**Documentation**
    https://sphobjinv.readthedocs.io/en/stable

**License**
    Code: `MIT License`_

    Docs & Docstrings: |CC BY 4.0|_

    See |license_txt|_ for full license terms.

**Members**

"""

import re

from sphobjinv.data import DataFields as DF  # noqa: N817
from sphobjinv.enum import HeaderFields as HF  # noqa: N817

#: Compiled |re| |bytes|  pattern for comment lines in decompressed
#: inventory files
pb_comments = re.compile(b"^#.*$", re.M)

#: Compiled |re| |bytes| pattern for project line
pb_project = re.compile(
    rf"""
    ^                            # Start of line
    [#][ ]Project:[ ]            # Preamble
    (?P<{HF.Project.value}>.*?)  # Lazy rest of line is project name
    \r?$                         # Ignore possible CR at EOL
    """.encode(encoding="utf-8"),
    re.M | re.X,
)

#: Compiled |re| |bytes| pattern for version line
pb_version = re.compile(
    rf"""
    ^                            # Start of line
    [#][ ]Version:[ ]            # Preamble
    (?P<{HF.Version.value}>.*?)  # Lazy rest of line is version
    \r?$                         # Ignore possible CR at EOL
    """.encode(encoding="utf-8"),
    re.M | re.X,
)

#: Regex pattern string used to compile
#: :data:`~sphobjinv.re.p_data` and
#: :data:`~sphobjinv.re.pb_data`
ptn_data = rf"""
    ^                               # Start of line
    (?P<{DF.Name.value}>.+?)        # --> Name
    \s+                             # Dividing space
    (?P<{DF.Domain.value}>[^\s:]+)  # --> Domain
    :                               # Dividing colon
    (?P<{DF.Role.value}>[^\s]+)     # --> Role
    \s+                             # Dividing space
    (?P<{DF.Priority.value}>-?\d+)  # --> Priority
    \s+?                            # Dividing space
    (?P<{DF.URI.value}>\S*)         # --> URI
    \s+                             # Dividing space
    (?P<{DF.DispName.value}>.+?)    # --> Display name, lazy b/c possible CR
    \r?$                            # Ignore possible CR at EOL
    """

#: Compiled |re| |bytes| regex pattern for data lines in |bytes| decompressed
#: inventory files
pb_data = re.compile(ptn_data.encode(encoding="utf-8"), re.M | re.X)

#: Compiled |re| |str| regex pattern for data lines in |str| decompressed
#: inventory files
p_data = re.compile(ptn_data, re.M | re.X)
